# -*- coding: utf-8 -*-
"""Monkey patches class"""

# pylint: disable=c0413, c0415, e0401, e1121
from django.db import DEFAULT_DB_ALIAS
from django.db import transaction


def django_sqlite_atomic():  # NOSONAR
    """monkey patch for django deployments fixing database lock issues"""

    def atomic(using: str = None, savepoint: bool = True, immediate: bool = False):
        # Bare decorator: @atomic -- although the first argument is called
        # `using`, it's actually the function being decorated.
        if callable(using):
            atomic_ = transaction.Atomic(DEFAULT_DB_ALIAS, savepoint, True)(using)
        # Decorator: @atomic(...) or context manager: with atomic(...): ...
        else:
            atomic_ = transaction.Atomic(using, savepoint, True)

        atomic_.immediate = immediate
        return atomic_

    def __enter__(self):
        """enter function"""
        connection = transaction.get_connection(self.using)
        if not connection.in_atomic_block:
            # Reset state when entering an outermost atomic block.
            connection.commit_on_exit = True
            connection.needs_rollback = False
            if not connection.get_autocommit():
                # Pretend we're already in an atomic block to bypass the code
                # that disables autocommit to enter a transaction, and make a
                # note to deal with this case in __exit__.
                connection.in_atomic_block = True
                connection.commit_on_exit = False

        if connection.in_atomic_block:
            # We're already in a transaction; create a savepoint, unless we
            # were told not to or we're already waiting for a rollback. The
            # second condition avoids creating useless savepoints and prevents
            # overwriting needs_rollback until the rollback is performed.
            if self.savepoint and not connection.needs_rollback:
                sid = connection.savepoint()
                connection.savepoint_ids.append(sid)
            else:
                connection.savepoint_ids.append(None)
        else:
            if self.immediate:
                connection.set_autocommit(False)
                connection.cursor().execute("BEGIN IMMEDIATE")

            else:
                connection.set_autocommit(
                    False, force_begin_transaction_with_broken_autocommit=True
                )

            connection.in_atomic_block = True

    transaction.atomic = atomic
    transaction.Atomic.immediate = False
    transaction.Atomic.__enter__ = __enter__


django_sqlite_atomic()



#!/usr/bin/python3
"""Verify and list EAB keyfile entries."""

from __future__ import annotations

import argparse
import os.path
from typing import Dict, Tuple

import yaml

from acme2certifier.acme_srv.db_handler import initialize
from acme2certifier.acme_srv.helper import (
    config_eab_profile_load,
    eab_handler_load,
    load_config,
    logger_setup,
    print_debug,
)


def _eab_dic_print(logger, eab_dic: Dict[str, str], config_dic: Dict[str, str]) -> None:
    """print eab dic"""
    logger.debug("eab_print_dic()")

    if config_dic["summary"] or config_dic["verbose"] or config_dic["veryverbose"]:
        _summary_print(logger, eab_dic)

    if config_dic["verbose"]:
        for key, value in eab_dic.items():
            if "hmac" in value:
                print(f'{key}: {value["hmac"]}')
            else:
                print(f"{key}: {value}")
    elif config_dic["veryverbose"]:
        print(yaml.dump(eab_dic, default_flow_style=False, default_style=""))


def _summary_print(logger, eab_dic: Dict[str, str]) -> None:
    """print summary of eab dic"""
    logger.debug("summary_print()")
    print(f"Summary: {len(eab_dic.keys())} entries in kid_file")


def _filter_eab_dic(logger, eab_dic: Dict[str, str], keyid: str) -> Dict[str, str]:
    """filter eab dic"""
    logger.debug("_filter_eab_dic(%s)", keyid)
    return {k: v for k, v in eab_dic.items() if k == keyid}


def arg_parse() -> Tuple[bool, Dict[str, str]]:
    """simple argparser"""
    parser = argparse.ArgumentParser(description="a2c-eab-chk - verify eab keyfile")
    parser.add_argument("-c", "--configfile", help="configfile", required=True)
    parser.add_argument(
        "-d", "--debug", help="debug mode", action="store_true", default=False
    )
    parser.add_argument(
        "-v", "--verbose", help="verbose", action="store_true", default=False
    )
    parser.add_argument(
        "-vv",
        "--veryverbose",
        help="show enrollment profile",
        action="store_true",
        default=False,
    )
    clist = parser.add_mutually_exclusive_group()
    clist.add_argument("-k", "--keyid", help="keyid to filter", default=None)
    clist.add_argument(
        "-s", "--summary", help="summary", default=False, action="store_true"
    )

    args = parser.parse_args()

    debug = args.debug
    config_dic = {
        "debug": args.debug,
        "verbose": args.verbose,
        "veryverbose": args.veryverbose,
        "keyid": args.keyid,
        "summary": args.summary,
        "configfile": args.configfile,
    }

    if not config_dic["summary"] and not config_dic["keyid"]:
        config_dic["summary"] = True

    return (debug, config_dic)


def eab_dic_load(logger, acme_srv_dic: Dict[str, Dict[str, str]]) -> Dict[str, str]:
    """load eabhandler"""
    logger.debug("eab_dic_load()")

    eab_profiling, eab_module = config_eab_profile_load(logger, acme_srv_dic)
    if not eab_profiling:
        eab_handler_module = eab_handler_load(logger, acme_srv_dic)
        eab_module = eab_handler_module.EABhandler

    with eab_module(logger) as eab_handler:
        eab_dic = eab_handler.key_file_load()

    return eab_dic


def main() -> None:
    """Load and print EAB keyfile contents according to CLI flags."""
    initialize()
    debug, config_dic = arg_parse()
    logger = logger_setup(debug)

    if os.path.exists(config_dic["configfile"]):
        acme_srv_dic = load_config(cfg_file=config_dic["configfile"])
    else:
        acme_srv_dic = {}
        error_text = f'Configfile {config_dic["configfile"]} not found.'
        logger.debug(error_text)
        print_debug(True, error_text)

    if "EABhandler" in acme_srv_dic:
        eab_dic = eab_dic_load(logger, acme_srv_dic)
        if config_dic.get("keyid"):
            eab_dic = _filter_eab_dic(logger, eab_dic, config_dic["keyid"])
    else:
        eab_dic = None
        print_debug(True, "No EABhandler section in configfile")

    if eab_dic:
        _eab_dic_print(logger, eab_dic, config_dic)


if __name__ == "__main__":
    main()



"""
Challenge Validator Registry.

Provides a registry system for managing and accessing challenge validators.
"""

from typing import Dict, List, Optional
import logging
from .base import (
    ChallengeValidator,
    ChallengeContext,
    ValidationResult,
    InvalidChallengeTypeError,
)


class ChallengeValidatorRegistry:
    """Registry for managing challenge validators."""

    def __init__(self, logger: logging.Logger):
        self.logger = logger
        self._validators: Dict[str, ChallengeValidator] = {}

    def register_validator(self, validator: ChallengeValidator) -> None:
        """Register a challenge validator."""
        self.logger.debug("ChallengeValidatorRegistry.register_validator()")
        challenge_type = validator.get_challenge_type()
        self._validators[challenge_type] = validator
        self.logger.debug(
            "ChallengeValidatorRegistry.register_validator(): Registered validator for challenge type: %s",
            challenge_type,
        )

    def get_validator(self, challenge_type: str) -> Optional[ChallengeValidator]:
        """Get a validator for the specified challenge type."""
        self.logger.debug(
            "ChallengeValidatorRegistry.get_validator(%s)", challenge_type
        )
        return self._validators.get(challenge_type)

    def get_supported_types(self) -> List[str]:
        """Get list of supported challenge types."""
        self.logger.debug("ChallengeValidatorRegistry.get_supported_types()")
        return list(self._validators.keys())

    def is_supported(self, challenge_type: str) -> bool:
        """Check if a challenge type is supported."""
        self.logger.debug("ChallengeValidatorRegistry.is_supported(%s)", challenge_type)
        return challenge_type in self._validators

    def validate_challenge(
        self, challenge_type: str, context: ChallengeContext
    ) -> ValidationResult:
        """Validate a challenge using the appropriate validator."""
        self.logger.debug(
            "ChallengeValidatorRegistry.validate_challenge(%s)", challenge_type
        )
        validator = self.get_validator(challenge_type)
        if not validator:
            raise InvalidChallengeTypeError(
                f"Unsupported challenge type: {challenge_type}"
            )

        return validator.validate_challenge(context)



# -*- coding: utf-8 -*-
"""ca handler for generic cmpv2 ca handler"""

from __future__ import print_function
import os
import shutil
import subprocess
import tempfile
from typing import List, Tuple, Dict

# pylint: disable=e0401
from acme2certifier.acme_srv.helper import (
    load_config,
    build_pem_file,
    b64_url_recode,
    config_profile_load,
)


class CAhandler(object):
    """EST CA  handler"""

    def __init__(self, _debug: bool = False, logger: object = None):
        self.logger = logger
        self.config_dic = {}
        self.openssl_bin = None
        self.tmp_dir = None
        self.recipient = None
        self.ref = None
        self.secret = None
        self.ca_pubs_file = None
        self.cert_file = None
        self.profiles = {}

    def __enter__(self):
        """Makes CAhandler a Context Manager"""
        if not self.openssl_bin:
            self._config_load()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _certs_bundle(self) -> Tuple[str, str]:
        """create needed cert(bundle)"""
        self.logger.debug("CAhandler._certs_bundle()")

        cert_raw = None
        cert_bundle = None
        ca_pem = None

        if os.path.isfile(self.ca_pubs_file):
            with open(self.ca_pubs_file, "r", encoding="utf-8") as fso:
                ca_pem = fso.read()

        # open certificate
        if os.path.isfile(self.cert_file):
            with open(self.cert_file, "r", encoding="utf-8") as fso:
                cert_raw = fso.read()

            # create bundle and raw cert
            if cert_raw and ca_pem:
                cert_bundle = cert_raw + ca_pem
            elif cert_raw:
                cert_bundle = cert_raw
            if cert_raw:
                cert_raw = cert_raw.replace("-----BEGIN CERTIFICATE-----\n", "")
                cert_raw = cert_raw.replace("-----END CERTIFICATE-----\n", "")
                cert_raw = cert_raw.replace("\n", "")
        self.logger.debug(
            "CAhandler._certs_bundle() ended with %s/%s",
            bool(cert_bundle),
            bool(cert_raw),
        )
        return (cert_bundle, cert_raw)

    def _config_refsecret_load(self, config_dic: Dict[str, str]):
        """ " load ref secrets from file"""
        self.logger.debug("CAhandler._config_refsecret_load()")

        if "CAhandler" in config_dic and "cmp_ref" in config_dic["CAhandler"]:
            if self.ref:
                self.logger.info("Overwrite cmp_ref variable")
            self.ref = config_dic["CAhandler"]["cmp_ref"]
        if "CAhandler" in config_dic and "cmp_secret" in config_dic["CAhandler"]:
            if self.secret:
                self.logger.info("Overwrite cmp_secret variable")
            self.secret = config_dic["CAhandler"]["cmp_secret"]

        self.logger.debug("CAhandler._config_refsecret_load() ended")

    def _config_paramters_load(self):
        """ " load refsecrets from file"""
        self.logger.debug("CAhandler._config_paramters_load()")

        if "cmd" not in self.config_dic:
            self.config_dic["cmd"] = "ir"
        if "popo" not in self.config_dic:
            self.config_dic["popo"] = 0

        # create temporary directory
        self.tmp_dir = tempfile.mkdtemp()
        self.ca_pubs_file = f"{self.tmp_dir}/capubs.pem"
        self.cert_file = f"{self.tmp_dir}/cert.pem"

        # defaulting openssl_bin
        if not self.openssl_bin:
            self.logger.warning(
                "cmp_openssl_bin parameter missing in configuration. Using default: /usr/bin/openssl"
            )
            self.openssl_bin = "/usr/bin/openssl"

        if not self.recipient:
            self.logger.error("cmp_recipient parameter missing in configuration.")

        self.logger.debug("CAhandler._config_paramters_load() ended")

    def _config_cmprecipient_load(self, config_dic: Dict[str, str]):
        """load and format recipient"""
        self.logger.debug("CAhandler._config_cmprecipient_load()")

        if config_dic["CAhandler"]["cmp_recipient"].startswith("/"):
            value = config_dic["CAhandler"]["cmp_recipient"]
        else:
            value = "/" + config_dic["CAhandler"]["cmp_recipient"]
        value = value.replace(", ", "/")
        value = value.replace(",", "/")
        self.config_dic["recipient"] = value

        self.logger.debug("CAhandler._config_cmprecipient_load() ended")

    def _config_cmpparameter_load(self, ele: str, config_dic: Dict[str, str]):
        """load cmp parameters"""
        self.logger.debug("CAhandler._config_cmpparameter_load()")

        if ele == "cmp_openssl_bin":
            self.openssl_bin = config_dic["CAhandler"]["cmp_openssl_bin"]
        elif ele == "cmp_recipient":
            self._config_cmprecipient_load(config_dic)
        elif ele == "cmp_ref_variable":
            try:
                self.ref = os.environ[config_dic["CAhandler"]["cmp_ref_variable"]]
            except Exception as err:
                self.logger.error("Could not load cmp_ref:%s", err)
        elif ele == "cmp_secret_variable":
            try:
                self.secret = os.environ[config_dic["CAhandler"]["cmp_secret_variable"]]
            except Exception as err:
                self.logger.error(
                    "Could not load cmp_secret_variable:%s",
                    err,
                )
        elif ele in ("cmp_secret", "cmp_ref"):
            self.logger.debug("CAhandler._config_cmpparameter_load() ignore %s", ele)
        else:
            if (
                config_dic["CAhandler"][ele] == "True"
                or config_dic["CAhandler"][ele] == "False"
            ):
                self.config_dic[ele[4:]] = config_dic.getboolean(
                    "CAhandler", ele, fallback=False
                )
            else:
                self.config_dic[ele[4:]] = config_dic["CAhandler"][ele]

        self.logger.debug("CAhandler._config_cmpparameter_load() ended")

    def _config_load(self):
        """ " load config from file"""
        # pylint: disable=R0912, R0915
        self.logger.debug("CAhandler._config_load()")
        config_dic = load_config(self.logger, "CAhandler")

        if "CAhandler" in config_dic:
            for ele in config_dic["CAhandler"]:
                if ele.startswith("cmp_"):
                    self._config_cmpparameter_load(ele, config_dic)

        # load ref/psk information
        self._config_refsecret_load(config_dic)
        # load file and directory names
        self._config_paramters_load()

        # load profiles
        self.profiles = config_profile_load(self.logger, config_dic)
        self.logger.debug("CAhandler._config_load() ended")

    def _opensslcmd_build(self) -> List[str]:
        """build openssl command"""
        self.logger.debug("CAhandler._opensslcmd_build()")

        cmd_list = [self.openssl_bin, "cmp"]
        for ele, value in self.config_dic.items():
            cmd_list.append(f"-{str(ele)}")
            if value is not True:
                cmd_list.append(str(value))

        cmd_list.extend(
            [
                "-csr",
                f"{self.tmp_dir}/csr.pem",
                "-extracertsout",
                self.ca_pubs_file,
                "-certout",
                self.cert_file,
            ]
        )

        # set timeouts if not configured
        if "-msg_timeout" not in cmd_list:
            cmd_list.extend(["-msg_timeout", "5"])
        if "-total_timeout" not in cmd_list:
            cmd_list.extend(["-total_timeout", "10"])

        if self.secret and self.ref:
            cmd_list.extend(["-ref", self.ref])
        if self.secret and self.ref:
            cmd_list.extend(["-secret", self.secret])

        self.logger.debug(
            "CAhandler._opensslcmd_build() ended with: %s", " ".join(cmd_list)
        )
        return cmd_list

    def _file_save(self, filename: str, content: str):
        """save content to file"""
        self.logger.debug("CAhandler._file_save(%s)", filename)
        with open(filename, "w", encoding="utf-8") as fso:
            fso.write(content)
        self.logger.debug("CAhandler._file_save() ended")

    def _tmp_dir_delete(self):
        """delete temp files"""
        self.logger.debug("CAhandler._tmp_dir_delete(%s)", self.tmp_dir)

        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)
        else:
            self.logger.error("Could not delate tmp_dir: %s", self.tmp_dir)

    def enroll(self, csr: str) -> Tuple[str, str, str, bool]:
        """enroll certificate from via MS certsrv"""
        self.logger.debug("CAhandler.enroll()")
        cert_bundle = None
        error = None
        cert_raw = None

        if self.openssl_bin:

            # prepare the CSR to be signed
            csr = build_pem_file(
                self.logger, None, b64_url_recode(self.logger, csr), None, True
            )
            # dump csr key
            self._file_save(f"{self.tmp_dir}/csr.pem", csr)

            # build openssl command and run it
            openssl_cmd = self._opensslcmd_build()
            rcode = subprocess.call(openssl_cmd)
            if rcode:
                self.logger.error(f"Enrollment failed with rcode: {rcode}")
                error = "rc from enrollment not 0"

            # generate certificates we need to return
            if os.path.isfile(f"{self.tmp_dir}/cert.pem"):
                cert_bundle, cert_raw = self._certs_bundle()
            else:
                error = "Enrollment failed"

            # delete temporary files
            self._tmp_dir_delete()

        else:
            error = "Config incomplete"

        self.logger.debug("Certificate.enroll() ended with error: %s", error)
        return (error, cert_bundle, cert_raw, None)

    def poll(
        self, _cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[str, str, str, str, str, bool]:
        """poll status of pending CSR and download certificates"""
        self.logger.debug("CAhandler.poll()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None
        rejected = False

        self.logger.debug("CAhandler.poll() ended")
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def revoke(
        self, _cert: str, _rev_reason: str, _rev_date: str
    ) -> Tuple[int, str, str]:
        """revoke certificate"""
        self.logger.debug("CAhandler.tsg_id_lookup()")

        # get serial from pem file and convert to formated hex

        code = 500
        message = "urn:ietf:params:acme:error:serverInternal"
        detail = "Revocation is not supported."

        return (code, message, detail)

    def trigger(self, _payload: str) -> Tuple[int, str, str]:
        """process trigger message and return certificate"""
        self.logger.debug("CAhandler.trigger()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)



# -*- coding: utf-8 -*-
"""EAB (External Account Binding) utilities for acme2certifier"""

import logging
from typing import Optional
from .csr import csr_subject_get
from .encoding import b64_url_recode
from .config import client_parameter_validate, profile_lookup, header_info_lookup
from .validation import cn_validate
from .domain_utils import allowed_domainlist_check


def _handle_eab_profiling(
    logger: logging.Logger, cahandler, csr: str, handler_hifield: str
) -> Optional[str]:
    """Handle EAB profiling logic"""
    logger.debug("Helper._handle_eab_profiling()")

    if not (hasattr(cahandler, "eab_handler") and cahandler.eab_handler):
        logger.error("EAB profiling enabled but no handler defined")
        return "Eab_profiling enabled but no handler defined"

    # profiling enabled - check profile
    return eab_profile_check(logger, cahandler, csr, handler_hifield)


def _handle_acme_profiling(
    logger: logging.Logger, cahandler, csr: str, handler_hifield: str
) -> None:
    """Handle ACME profiling logic"""
    logger.debug("Helper._handle_acme_profiling()")

    profile = profile_lookup(logger, csr)
    if profile:
        logger.debug(
            "Helper.profile_lookup(): setting %s to %s",
            handler_hifield,
            profile,
        )
        setattr(cahandler, handler_hifield, profile)


def _handle_header_info_profiling(
    logger: logging.Logger, cahandler, csr: str, handler_hifield: str
) -> None:
    """Handle header info profiling logic"""
    logger.debug("Helper._handle_header_info_profiling()")

    hil_value = header_info_lookup(
        logger, csr, cahandler.header_info_field, handler_hifield
    )
    if hil_value:
        logger.debug(
            "Helper.eab_profile_header_info_check(): setting %s to %s",
            handler_hifield,
            hil_value,
        )
        logger.info(
            "Received enrollment parameter: %s value: %s via headerinfo field",
            handler_hifield,
            hil_value,
        )
        setattr(cahandler, handler_hifield, hil_value)
    else:
        logger.debug("eab_profile_header_info_check(): no header_info field found")


def eab_profile_header_info_check(
    logger: logging.Logger,
    cahandler,
    csr: str,
    handler_hifield: str = "profile_name",
) -> Optional[str]:
    """check profile"""
    logger.debug("Helper.eab_profile_header_info_check()")

    # Priority 1: EAB profiling - returns error string or None
    if hasattr(cahandler, "eab_profiling") and cahandler.eab_profiling:
        error = _handle_eab_profiling(logger, cahandler, csr, handler_hifield)

    # Priority 2: ACME profiling (preferred over header info) - no errors
    elif hasattr(cahandler, "profiles") and cahandler.profiles:
        _handle_acme_profiling(logger, cahandler, csr, handler_hifield)
        error = None

    # Priority 3: Header info profiling - no errors
    elif hasattr(cahandler, "header_info_field") and cahandler.header_info_field:
        _handle_header_info_profiling(logger, cahandler, csr, handler_hifield)
        error = None

    # Priority 4: No profiling
    else:
        error = None

    logger.debug("Helper.eab_profile_header_info_check() ended with %s", error)
    return error


def eab_profile_subject_string_check(
    logger: logging.Logger, profile_subject_dic, key: str, value: str
) -> str:
    """check if a for a string value taken from profile if its a variable inside a class and apply value"""
    logger.debug(
        "Helper.eab_profile_subject_string_check(): string: key: %s, value: %s",
        key,
        value,
    )

    error = False
    if key == "commonName":
        # check if CN is a valid IP address or fqdn
        error = cn_validate(logger, value)
    elif key in profile_subject_dic:
        if isinstance(profile_subject_dic[key], str) and (
            value == profile_subject_dic[key] or profile_subject_dic[key] == "*"
        ):
            logger.debug(
                "Helper.eab_profile_subject_check() successul for string : %s", key
            )
            del profile_subject_dic[key]
        elif (
            isinstance(profile_subject_dic[key], list)
            and value in profile_subject_dic[key]
        ):
            logger.debug(
                "Helper.eab_profile_subject_check() successul for list : %s", key
            )
            del profile_subject_dic[key]
        else:
            logger.error(
                "EAB profile subject check failed for: %s: value: %s expected: %s",
                key,
                value,
                profile_subject_dic[key],
            )
            error = f"Profile subject check failed for {key}"
    else:
        logger.error("EAB profile subject failed for: %s", key)
        error = f"Profile subject check failed for {key}"

    logger.debug("Helper.eab_profile_subject_string_check() ended")
    return error


def eab_profile_subject_check(
    logger: logging.Logger, csr: str, profile_subject_dic: str
) -> str:
    """check subject against profile information"""
    logger.debug("Helper.eab_profile_subject_check()")
    error = None

    # get subject from csr
    subject_dic = csr_subject_get(logger, csr)

    # check if all profile subject entries are in csr
    for key, value in subject_dic.items():
        error = eab_profile_subject_string_check(
            logger, profile_subject_dic, key, value
        )
        if error:
            break

    # check if we have any entries left in the profile_subject_dic
    if not error and profile_subject_dic:
        logger.error(
            "EAB profile subject check failed for: %s",
            list(profile_subject_dic.keys()),
        )
        error = "Profile subject check failed"

    logger.debug("Helper.eab_profile_subject_check() ended with: %s", error)
    return error


def eab_profile_revocation_check(
    logger: logging.Logger, cahandler, certificate_raw: str
):
    """check eab profile for revocation"""
    logger.debug("Helper.eab_profile_revocation_check()")
    with cahandler.eab_handler(logger) as eab_handler:
        eab_profile_dic = eab_handler.eab_profile_get(
            b64_url_recode(logger, certificate_raw), revocation=True
        )
        for key, value in eab_profile_dic.items():
            if key in ["subject", "allowed_domainlist"]:
                continue
            elif isinstance(value, str):
                eab_profile_string_check(logger, cahandler, key, value)
            elif isinstance(value, list):
                # check if we need to execute a function from the handler
                if "eab_profile_list_check" in dir(cahandler):
                    _result = cahandler.eab_profile_list_check(
                        eab_handler, certificate_raw, key, value
                    )
                else:
                    _result = eab_profile_list_check(
                        logger, cahandler, eab_handler, certificate_raw, key, value
                    )

    logger.debug("Helper.eab_profile_revocation_check() ended")


def eab_profile_check(
    logger: logging.Logger, cahandler, csr: str, handler_hifield: str
) -> str:
    """check eab profile"""
    logger.debug("Helper.eab_profile_check()")

    result = None
    with cahandler.eab_handler(logger) as eab_handler:
        eab_profile_dic = eab_handler.eab_profile_get(csr)
        for key, value in eab_profile_dic.items():
            if key == "subject":
                result = eab_profile_subject_check(logger, csr, value)
            elif isinstance(value, str):
                eab_profile_string_check(logger, cahandler, key, value)
            elif isinstance(value, list):
                # check if we need to execute a function from the handler
                if "eab_profile_list_check" in dir(cahandler):
                    result = cahandler.eab_profile_list_check(
                        eab_handler, csr, key, value
                    )
                else:
                    result = eab_profile_list_check(
                        logger, cahandler, eab_handler, csr, key, value
                    )
            if result:
                break

        # we need to reject situations where profiling is enabled but the header_hifiled is not defined in json
        if cahandler.header_info_field and handler_hifield not in eab_profile_dic:
            hil_value = header_info_lookup(
                logger, csr, cahandler.header_info_field, handler_hifield
            )
            if hil_value:
                # setattr(self, handler_hifield, hil_value)
                result = (
                    f'header_info field "{handler_hifield}" is not allowed by profile'
                )

    logger.debug("Helper.eab_profile_check() ended with: %s", result)
    return result


def eab_profile_list_check(logger, cahandler, eab_handler, csr, key, value):
    """check if a for a list value taken from profile if its a variable inside a class and apply value"""
    logger.debug(
        "Helper.eab_profile_list_check(): list: key: %s, value: %s", key, value
    )

    result = None
    if hasattr(cahandler, key) and key != "allowed_domainlist":
        new_value, error = client_parameter_validate(logger, csr, cahandler, key, value)
        if new_value:
            logger.debug(
                "Helper.eab_profile_list_check(): setting attribute: %s to %s",
                key,
                new_value,
            )
            setattr(cahandler, key, new_value)
        else:
            result = error
    elif key == "allowed_domainlist":
        # check if csr contains allowed domains
        if "allowed_domains_check" in dir(eab_handler):
            # execute a function from eab_handler
            logger.info("Execute allowed_domains_check() from eab handler")
            error = eab_handler.allowed_domains_check(csr, value)
        else:
            # execute default adl function from helper
            logger.debug(
                "Helper.eab_profile_list_check(): execute default allowed_domainlist_check()"
            )
            error = allowed_domainlist_check(logger, csr, value)
        if error:
            result = error
    else:
        logger.warning(
            "EAP profile list checking: ignoring unrecognized list attribute: key: %s value: %s",
            key,
            value,
        )

    logger.debug("Helper.eab_profile_list_check() ended with: %s", result)
    return result


def eab_profile_string_check(logger, cahandler, key, value):
    """check if a for a string value taken from profile if its a variable inside a class and apply value"""
    logger.debug(
        "Helper.eab_profile_string_check(): string: key: %s, value: %s", key, value
    )

    if hasattr(cahandler, key):
        logger.debug(
            "Helper.eab_profile_string_check(): setting attribute: %s to %s", key, value
        )
        setattr(cahandler, key, value)
    else:
        logger.warning(
            "EAB profile string checking: ignoring unrecognized string attribute: key: %s value: %s",
            key,
            value,
        )

    logger.debug("Helper.eab_profile_string_check() ended")



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for acme2certifier"""

# pylint: disable= C0415, W0212
import unittest
import sys
import os
from unittest.mock import patch, Mock
import requests
import configparser

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class TestACMEHandler(unittest.TestCase):
    """test class for cgi_handler"""

    def setUp(self):
        """setup unittest"""
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.cahandlers.nclm_ca_handler import CAhandler

        self.cahandler = CAhandler(False, self.logger)
        # self.cahandler.api_host = 'api_host'
        # self.cahandler.auth = 'auth'

    def test_001_default(self):
        """default test which always passes"""
        self.assertEqual("foo", "foo")

    @patch.object(requests, "post")
    def test_002__api_post(self, mock_req):
        """test _api_post successful run"""
        mockresponse = Mock()
        mock_req.return_value = mockresponse
        mockresponse.json = lambda: {"foo": "bar"}
        self.assertEqual({"foo": "bar"}, self.cahandler._api_post("url", "data"))

    @patch("requests.post")
    def test_003__api_post(self, mock_post):
        """CAhandler.get_ca() returns an http error"""
        self.cahandler.api_host = "api_host"
        self.cahandler.auth = "auth"
        mock_post.side_effect = Exception("exc_api_post")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual("exc_api_post", self.cahandler._api_post("url", "data"))
        self.assertIn(
            "ERROR:test_a2c:API POST request failed: exc_api_post",
            lcm.output,
        )

    @patch.object(requests, "post")
    def test_004__api_post(self, mock_req):
        """test _api_post successful run"""
        mockresponse = Mock()
        mock_req.return_value = mockresponse
        mockresponse.status_code = "status_code"
        mockresponse.json = Exception("json_exc")
        self.assertEqual(
            {"status": "status_code"}, self.cahandler._api_post("url", "data")
        )

    def test_005__config_check(self):
        """CAhandler._config.check() no api_host"""
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("api_host to be set in config file", self.cahandler.error)
        self.assertIn('ERROR:test_a2c:Missing "api_host" in configuration.', lcm.output)

    def test_006__config_check(self):
        """CAhandler._config.check() no api_user"""
        self.cahandler.api_host = "api_host"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("api_user to be set in config file", self.cahandler.error)
        self.assertIn('ERROR:test_a2c:Missing "api_user" in configuration.', lcm.output)

    def test_007__config_check(self):
        """CAhandler._config.check() no api_user"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {"api_user": False}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("api_user to be set in config file", self.cahandler.error)
        self.assertIn('ERROR:test_a2c:Missing "api_user" in configuration.', lcm.output)

    def test_008__config_check(self):
        """CAhandler._config.check() no api_password"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {"api_user": "api_user"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("api_password to be set in config file", self.cahandler.error)
        self.assertIn(
            'ERROR:test_a2c:Missing "api_password" in configuration.', lcm.output
        )

    def test_009__config_check(self):
        """CAhandler._config.check() no api_password"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {"api_user": "api_user", "api_password": False}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("api_password to be set in config file", self.cahandler.error)
        self.assertIn(
            'ERROR:test_a2c:Missing "api_password" in configuration.', lcm.output
        )

    def test_010__config_check(self):
        """CAhandler._config.check() no tsg_name"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {
            "api_user": "api_user",
            "api_password": "api_password",
        }
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("tsg_name to be set in config file", self.cahandler.error)
        self.assertIn('ERROR:test_a2c:"tsg_name" to be set in config file', lcm.output)

    def test_011__config_check(self):
        """CAhandler._config.check() no tsg_name"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {
            "api_user": "api_user",
            "api_password": "api_password",
        }
        self.cahandler.container_info_dic = {"name": False}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("tsg_name to be set in config file", self.cahandler.error)
        self.assertIn('ERROR:test_a2c:"tsg_name" to be set in config file', lcm.output)

    def test_012__config_check(self):
        """CAhandler._config.check() no ca_name"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {
            "api_user": "api_user",
            "api_password": "api_password",
        }
        self.cahandler.container_info_dic = {"name": "name"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertEqual("ca_name to be set in config file", self.cahandler.error)
        self.assertIn('ERROR:test_a2c:"ca_name" to be set in config file', lcm.output)

    def test_013__config_check(self):
        """CAhandler._config.check() ca_bundle False"""
        self.cahandler.api_host = "api_host"
        self.cahandler.credential_dic = {
            "api_user": "api_user",
            "api_password": "api_password",
        }
        self.cahandler.container_info_dic = {"name": "name"}
        self.cahandler.ca_name = "ca_name"
        self.cahandler.ca_id_list = ["id1", "id2"]
        self.cahandler.ca_bundle = False
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_check()
        self.assertFalse(self.cahandler.error)
        self.assertIn(
            'WARNING:test_a2c:CA bundle validation is disabled ("ca_bundle" set to False). Server certificate will not be validated.',
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_014_config_load(self, mock_load_cfg):
        """CAhandler._config_load no cahandler section"""
        parser = configparser.ConfigParser()
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertTrue(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_015_config_load(self, mock_load_cfg):
        """CAhandler._config_load api_host"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_host": "api_host", "foo": "bar"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual("api_host", self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertTrue(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_016_config_load(self, mock_load_cfg):
        """CAhandler._config_load api_user"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_user": "api_user"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": "api_user", "api_password": None},
            self.cahandler.credential_dic,
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertTrue(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_017_config_load(self, mock_load_cfg):
        """CAhandler._config_load api_password"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_password": "api_password"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": "api_password"},
            self.cahandler.credential_dic,
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertTrue(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_018_config_load(self, mock_load_cfg):
        """CAhandler._config_load ca_name"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"ca_name": "ca_name"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertEqual("ca_name", self.cahandler.ca_name)
        self.assertTrue(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_019_config_load(self, mock_load_cfg):
        """CAhandler._config_load tsg_name"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"tsg_name": "tsg_name"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertIn(
            "WARNING:test_a2c:Configuration uses deprecated 'tsg_name'. Use 'container_name' instead.",
            lcm.output,
        )
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertTrue(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_020_config_load(self, mock_load_cfg):
        """CAhandler._config_load ca_bundle string"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"ca_bundle": "ca_bundle"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertEqual("ca_bundle", self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_021_config_load(self, mock_load_cfg):
        """CAhandler._config_load ca_bundle False"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"ca_bundle": False}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertEqual({"name": None, "id": None}, self.cahandler.template_info_dic)
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_022_config_load(self, mock_load_cfg):
        """CAhandler._config_load template_name"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"template_name": "template_name"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertFalse(self.cahandler.ca_name)
        self.assertEqual(
            {"name": "template_name", "id": None}, self.cahandler.template_info_dic
        )
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch.dict("os.environ", {"api_user_var": "user_var"})
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_023_config_load(self, mock_load_cfg):
        """CAhandler._config_load load username from variable"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_user_variable": "api_user_var"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": "user_var", "api_password": None},
            self.cahandler.credential_dic,
        )

    @patch.dict("os.environ", {"api_user_var": "user_var"})
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_024_config_load(self, mock_load_cfg):
        """CAhandler._config_load load username from non existing"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_user_variable": "does_not_exist"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertIn(
            "ERROR:test_a2c:Unable to load API user from environment: 'does_not_exist'",
            lcm.output,
        )

    @patch.dict("os.environ", {"api_user_var": "user_var"})
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_025_config_load(self, mock_load_cfg):
        """CAhandler._config_load load username from wich gets overwritten from cfg-file"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {
            "api_user_variable": "api_user_var",
            "api_user": "api_user",
        }
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": "api_user", "api_password": None},
            self.cahandler.credential_dic,
        )
        self.assertIn("INFO:test_a2c:Overwrite api_user", lcm.output)

    @patch.dict("os.environ", {"api_password_var": "password_var"})
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_026_config_load(self, mock_load_cfg):
        """CAhandler._config_load load password from variable"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_password_variable": "api_password_var"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": "password_var"},
            self.cahandler.credential_dic,
        )

    @patch.dict("os.environ", {"api_password_var": "password_var"})
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_027_config_load(self, mock_load_cfg):
        """CAhandler._config_load load password from non existing variable"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"api_password_variable": "does_not_exist"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_user": None, "api_password": None}, self.cahandler.credential_dic
        )
        self.assertIn(
            "ERROR:test_a2c:Could not load password_variable:'does_not_exist'",
            lcm.output,
        )

    @patch.dict("os.environ", {"api_password_var": "password_var"})
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_028_config_load(self, mock_load_cfg):
        """CAhandler._config_load load password from variable which gets overwritten"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {
            "api_password_variable": "api_password_var",
            "api_password": "api_password",
        }
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.api_host)
        self.assertEqual(
            {"api_password": "api_password", "api_user": None},
            self.cahandler.credential_dic,
        )
        self.assertIn("INFO:test_a2c:Overwrite api_password", lcm.output)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.parse_url")
    @patch("json.loads")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_029_config_load(self, mock_load_cfg, mock_json, mock_url):
        """test _config_load ca_handler configured load proxies"""
        parser = configparser.ConfigParser()
        parser["DEFAULT"] = {"proxy_server_list": "foo"}
        mock_load_cfg.return_value = parser
        mock_url.return_value = {"foo": "bar"}
        mock_json.return_value = "foo"
        self.cahandler._config_load()
        self.assertTrue(mock_json.called)
        self.assertTrue(mock_url.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.proxy_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.parse_url")
    @patch("json.loads")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_030_config_load(self, mock_load_cfg, mock_json, mock_url, mock_chk):
        """test _config_load ca_handler configured load proxies"""
        parser = configparser.ConfigParser()
        parser["DEFAULT"] = {"proxy_server_list": "foo"}
        mock_load_cfg.return_value = parser
        mock_url.return_value = {"host": "bar:8888"}
        mock_json.return_value = "foo.bar.local"
        mock_chk.return_value = "proxy.bar.local"
        self.cahandler._config_load()
        self.assertTrue(mock_json.called)
        self.assertTrue(mock_url.called)
        self.assertTrue(mock_chk.called)
        self.assertEqual(
            {"http": "proxy.bar.local", "https": "proxy.bar.local"},
            self.cahandler.proxy,
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.proxy_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.parse_url")
    @patch("json.loads")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_031_config_load(self, mock_load_cfg, mock_json, mock_url, mock_chk):
        """test _config_load ca_handler configured load proxies"""
        parser = configparser.ConfigParser()
        parser["DEFAULT"] = {"proxy_server_list": "foo"}
        mock_load_cfg.return_value = parser
        mock_url.return_value = {"host": "bar"}
        mock_json.return_value = "foo.bar.local"
        mock_chk.return_value = "proxy.bar.local"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertTrue(mock_json.called)
        self.assertTrue(mock_url.called)
        self.assertFalse(mock_chk.called)
        self.assertFalse(self.cahandler.proxy)
        self.assertIn(
            "WARNING:test_a2c:Failed to load proxy_server_list from configuration: not enough values to unpack (expected 2, got 1)",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_032_config_load(self, mock_load_cfg):
        """CAhandler._config_load request_delta_treshold"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"request_timeout": 10}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual(10, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_033_config_load(self, mock_load_cfg):
        """CAhandler._config_load request_delta_treshold"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"request_timeout": "aa"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual(20, self.cahandler.request_timeout)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.load_config")
    def test_034_config_load(self, mock_load_cfg):
        """CAhandler._config_load"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"container_name": "container_name"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual(
            {"name": "container_name", "id": None}, self.cahandler.container_info_dic
        )

    @patch("requests.post")
    @patch("requests.get")
    def test_035__login(self, mock_get, mock_post):
        """CAhandler._unusedrequests_get"""
        self.cahandler.api_host = "api_host"
        mockresponse1 = Mock()
        mockresponse1.status_code = "500"
        mockresponse1.ok = None
        mock_get.return_value = mockresponse1
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._login()
        self.assertIn("ERROR:test_a2c:Login failed. Error: 500", lcm.output)
        self.assertFalse(mock_post.called)
        self.assertFalse(self.cahandler.headers)

    @patch("requests.post")
    @patch("requests.get")
    def test_036__login(self, mock_get, mock_post):
        """CAhandler._unusedrequests_get"""
        self.cahandler.api_host = "api_host"
        mockresponse1 = Mock()
        mockresponse1.status_code = "200"
        mockresponse1.json = lambda: {"versionNumber": "versionNumber"}
        mockresponse1.ok = True
        mock_get.return_value = mockresponse1
        mockresponse2 = Mock()
        mockresponse2.status_code = "500"
        mockresponse2.json = lambda: {
            "foo": "bar",
            "username": "username",
            "realms": "realms",
        }
        mockresponse2.ok = None
        mock_post.return_value = mockresponse2
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._login()
        self.assertIn("ERROR:test_a2c:Login Error: 500", lcm.output)
        self.assertTrue(mock_post.called)
        self.assertFalse(self.cahandler.headers)
        self.assertEqual("versionNumber", self.cahandler.nclm_version)

    @patch("requests.post")
    @patch("requests.get")
    def test_037__login(self, mock_get, mock_post):
        """CAhandler._unusedrequests_get"""
        self.cahandler.api_host = "api_host"
        mockresponse1 = Mock()
        mockresponse1.status_code = "200"
        mockresponse1.json = lambda: {"versionNumber": "versionNumber"}
        mockresponse1.ok = True
        mock_get.return_value = mockresponse1
        mockresponse2 = Mock()
        mockresponse2.status_code = "200"
        mockresponse2.json = lambda: {
            "foo": "bar",
            "username": "username",
            "realms": "realms",
        }
        mockresponse2.ok = True
        mock_post.return_value = mockresponse2
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._login()
        self.assertIn(
            "ERROR:test_a2c:No token returned after logging in. Aborting.",
            lcm.output,
        )
        self.assertTrue(mock_post.called)
        self.assertFalse(self.cahandler.headers)
        self.assertEqual("versionNumber", self.cahandler.nclm_version)

    @patch("requests.post")
    @patch("requests.get")
    def test_038__login(self, mock_get, mock_post):
        """CAhandler._unusedrequests_get"""
        self.cahandler.api_host = "api_host"
        mockresponse1 = Mock()
        mockresponse1.status_code = "200"
        mockresponse1.json = lambda: {"versionNumber": "versionNumber"}
        mockresponse1.ok = True
        mock_get.return_value = mockresponse1
        mockresponse2 = Mock()
        mockresponse2.status_code = "200"
        mockresponse2.json = lambda: {
            "access_token": "access_token",
            "username": "username",
            "realms": "realms",
        }
        mockresponse2.ok = True
        mock_post.return_value = mockresponse2
        self.cahandler._login()
        self.assertTrue(mock_post.called)
        self.assertEqual(
            {"Authorization": "Bearer access_token"}, self.cahandler.headers
        )
        self.assertEqual("versionNumber", self.cahandler.nclm_version)

    def test_039_poll(self):
        """CAhandler.poll()"""
        self.assertEqual(
            ("Method not implemented.", None, None, "poll_identifier", False),
            self.cahandler.poll("cert_name", "poll_identifier", "csr"),
        )

    def test_040_trigger(self):
        """CAhandler.trigger()"""
        self.assertEqual(
            ("Method not implemented.", None, None), self.cahandler.trigger("payload")
        )

    @patch("requests.get")
    def test_041_container_id_lookup(self, mock_get):
        """CAhandler._container_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {
            "items": [{"name": "name1", "id": "id1"}, {"name": "name2", "id": "id2"}]
        }
        mock_get.return_value = mockresponse
        self.cahandler.container_info_dic = {"name": "name1", "id": None}
        self.cahandler._container_id_lookup()
        self.assertEqual(
            {"name": "name1", "id": "id1"}, self.cahandler.container_info_dic
        )

    @patch("requests.get")
    def test_042_container_id_lookup(self, mock_get):
        """CAhandler._container_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {
            "items": [{"name1": "name1", "id": "id1"}, {"name": "name2", "id": "id2"}]
        }
        mock_get.return_value = mockresponse
        self.cahandler.container_info_dic = {"name": "name", "id": None}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._container_id_lookup()
        self.assertIn(
            "ERROR:test_a2c:Incomplete container response: {'name1': 'name1', 'id': 'id1'}",
            lcm.output,
        )
        self.assertEqual(
            {"name": "name", "id": None}, self.cahandler.container_info_dic
        )

    @patch("requests.get")
    def test_043_container_id_lookup(self, mock_get):
        """CAhandler._container_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {
            "foo": [{"name": "name1", "id": "id1"}, {"name": "name2", "id": "id2"}]
        }
        mock_get.return_value = mockresponse
        self.cahandler.container_info_dic = {"name": "name", "id": None}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._container_id_lookup()
        self.assertIn(
            "ERROR:test_a2c:No target system groups found for filter: name.",
            lcm.output,
        )
        self.assertEqual(
            {"name": "name", "id": None}, self.cahandler.container_info_dic
        )

    @patch("requests.get")
    def test_044_container_id_lookup(self, mock_req):
        """CAhandler._container_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mock_req.side_effect = Exception("exc_container_id_lookup")
        self.cahandler.container_info_dic = {"name": "name", "id": None}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._container_id_lookup()
        self.assertIn(
            "ERROR:test_a2c:Failed to retrieve container id: exc_container_id_lookup",
            lcm.output,
        )
        self.assertIn(
            "ERROR:test_a2c:No target system groups found for filter: name.",
            lcm.output,
        )
        self.assertEqual(
            {"name": "name", "id": None}, self.cahandler.container_info_dic
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._templates_enumerate")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_list_get")
    def test_045__template_id_lookup(self, mock_list, mock_enum):
        """CAhandler._template_id_lookup"""
        mock_list.return_value = {"foo": "bar"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._template_id_lookup("caid")
        self.assertIn(
            "ERROR:test_a2c:No templates found for filter: None.",
            lcm.output,
        )
        self.assertFalse(mock_enum.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._templates_enumerate")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_list_get")
    def test_046__template_id_lookup(self, mock_list, mock_enum):
        """CAhandler._template_id_lookup"""
        mock_list.return_value = {"items": ["foo", "bar"]}
        self.cahandler._template_id_lookup("caid")
        self.assertTrue(mock_enum.called)

    @patch("requests.get")
    def test_047__template_list_get(self, mock_get):
        """CAhandler._template_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {"foo": "bar"}
        mock_get.return_value = mockresponse
        self.assertEqual({"foo": "bar"}, self.cahandler._template_list_get(6))

    @patch("requests.get")
    def test_048__template_list_get(self, mock_get):
        """CAhandler._template_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mock_get.side_effect = Exception("req_exc")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.cahandler._template_list_get(6))
        self.assertIn(
            "ERROR:test_a2c:Failed to retrieve template list: req_exc",
            lcm.output,
        )

    @patch("requests.get")
    def test_049__template_list_get(self, mock_get):
        """CAhandler._template_id_lookup()"""
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {"items": "bar"}
        mock_get.return_value = mockresponse
        self.assertEqual({"items": "bar"}, self.cahandler._template_list_get(6))

    def test_050__templates_enumerate(self):
        """CAhandler._templates_enumerate()"""
        template_list = {
            "items": [{"name": "foo", "id": "id"}, {"name": "foo1", "id": "id1"}]
        }
        self.cahandler.template_info_dic = {"name": "foo"}
        self.cahandler._templates_enumerate(template_list)
        self.assertEqual({"id": "id", "name": "foo"}, self.cahandler.template_info_dic)

    def test_051__templates_enumerate(self):
        """CAhandler._templates_enumerate()"""
        template_list = {
            "items": [{"name": "foo", "id": "id"}, {"name": "foo1", "id": "id1"}]
        }
        self.cahandler.template_info_dic = {"name": "foo1"}
        self.cahandler._templates_enumerate(template_list)
        self.assertEqual(
            {"id": "id1", "name": "foo1"}, self.cahandler.template_info_dic
        )

    def test_052__templates_enumerate(self):
        """CAhandler._templates_enumerate()"""
        template_list = {
            "items": [
                {"name": "foo", "id": "id"},
                {"name": "foo1", "id": "id1"},
                {"name": "foo", "id": "id2"},
            ]
        }
        self.cahandler.template_info_dic = {"name": "foo"}
        self.cahandler._templates_enumerate(template_list)
        self.assertEqual({"id": "id", "name": "foo"}, self.cahandler.template_info_dic)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_load")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._login")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._container_id_lookup")
    def test_053__enter__(self, mock_lookup, mock_login, mock_check, mock_load):
        """test enter"""
        self.cahandler.__enter__()
        self.assertTrue(mock_load.called)
        self.assertTrue(mock_check.called)
        self.assertTrue(mock_login.called)
        self.assertTrue(mock_lookup.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_load")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._login")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._container_id_lookup")
    def test_054__enter__(self, mock_lookup, mock_login, mock_check, mock_load):
        """test enter  with host already defined"""
        self.cahandler.api_host = "api_host"
        self.cahandler.__enter__()
        self.assertFalse(mock_load.called)
        self.assertFalse(mock_check.called)
        self.assertTrue(mock_login.called)
        self.assertTrue(mock_lookup.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_load")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._login")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._container_id_lookup")
    def test_055__enter__(self, mock_lookup, mock_login, mock_check, mock_load):
        """test enter with header defined"""
        self.cahandler.headers = "header"
        self.cahandler.__enter__()
        self.assertTrue(mock_load.called)
        self.assertTrue(mock_check.called)
        self.assertFalse(mock_login.called)
        self.assertTrue(mock_lookup.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_load")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._login")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._container_id_lookup")
    def test_056__enter__(self, mock_lookup, mock_login, mock_check, mock_load):
        """test enter with error defined"""
        self.cahandler.error = "error"
        self.cahandler.__enter__()
        self.assertTrue(mock_load.called)
        self.assertTrue(mock_check.called)
        self.assertFalse(mock_login.called)
        self.assertFalse(mock_lookup.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_load")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._login")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._container_id_lookup")
    def test_057__enter__(self, mock_lookup, mock_login, mock_check, mock_load):
        """test enter with tst_info_dic defined"""
        self.cahandler.container_info_dic = {"id": "foo"}
        self.cahandler.__enter__()
        self.assertTrue(mock_load.called)
        self.assertTrue(mock_check.called)
        self.assertTrue(mock_login.called)
        self.assertFalse(mock_lookup.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_load")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._login")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._container_id_lookup")
    def test_058__enter__(self, mock_lookup, mock_login, mock_check, mock_load):
        """test enter with error defined"""
        self.cahandler.container_info_dic = {"id": "foo"}
        self.cahandler.error = "error"
        self.cahandler.__enter__()
        self.assertTrue(mock_load.called)
        self.assertTrue(mock_check.called)
        self.assertFalse(mock_login.called)
        self.assertFalse(mock_lookup.called)

    def test_059__ca_id_get(self):
        """test _ca_id_get()"""
        ca_list = {}
        self.assertFalse(self.cahandler._ca_id_get(ca_list))

    def test_060__ca_id_get(self):
        """test _ca_id_get()"""
        ca_list = {"ca": {"foo": "bar"}}
        self.assertFalse(self.cahandler._ca_id_get(ca_list))

    def test_061__ca_id_get(self):
        """test _ca_id_get()"""
        ca_list = {"items": "bar"}
        self.assertFalse(self.cahandler._ca_id_get(ca_list))

    def test_062__ca_id_get(self):
        """test _ca_id_get()"""
        ca_list = {"items": [{"foo": "bar"}]}
        self.assertFalse(self.cahandler._ca_id_get(ca_list))

    def test_063__ca_id_get(self):
        """test _ca_id_get()"""
        self.cahandler.ca_name = "ca_name"
        ca_list = {"items": [{"name": "ca_name", "id": "id"}]}
        self.assertEqual("id", self.cahandler._ca_id_get(ca_list))

    def test_064__ca_id_get(self):
        """test _ca_id_get()"""
        self.cahandler.ca_name = "ca_name"
        ca_list = {"items": [{"name": "ca_name", "id1": "id"}]}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.cahandler._ca_id_get(ca_list))
        self.assertIn(
            "ERROR:test_a2c:CA response missing policyLinkId field.",
            lcm.output,
        )

    def test_065__ca_id_get(self):
        """test _ca_id_get()"""
        self.cahandler.ca_name = "ca_name1"
        ca_list = {"items": [{"name": "ca_name", "id": "id"}]}
        self.assertFalse(self.cahandler._ca_id_get(ca_list))

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_id_get")
    @patch("requests.get")
    def test_066__ca_policylink_id_lookup(self, mock_req, mock_caid):
        """test _ca_policylink_id_lookup()"""
        self.cahandler.api_host = "api_host"
        self.cahandler.container_info_dic = {"id": "id"}
        mockresponse = Mock()
        mockresponse.json = lambda: {"items": ["foo", "bar", "foo", "bar"]}
        mock_req.return_value = mockresponse
        mock_caid.return_value = 10
        self.assertEqual(10, self.cahandler._ca_policylink_id_lookup())
        self.assertTrue(mock_caid.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_id_get")
    @patch("requests.get")
    def test_067__ca_policylink_id_lookup(self, mock_req, mock_caid):
        """test _ca_policylink_id_lookup()"""
        self.cahandler.api_host = "api_host"
        self.cahandler.container_info_dic = {"id": "id"}
        mockresponse = Mock()
        mockresponse.json = lambda: {"items": ["foo", "bar", "foo", "bar"]}
        mock_req.return_value = mockresponse
        mock_caid.return_value = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.cahandler._ca_policylink_id_lookup())
        self.assertIn(
            "ERROR:test_a2c:No policy link ID found for CA name: None",
            lcm.output,
        )
        self.assertTrue(mock_caid.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_id_get")
    @patch("requests.get")
    def test_068__ca_policylink_id_lookup(self, mock_req, mock_caid):
        """test _ca_policylink_id_lookup()"""
        self.cahandler.api_host = "api_host"
        self.cahandler.container_info_dic = {"id": "id"}
        mockresponse = Mock()
        mockresponse.json = lambda: {"foo": ["foo", "bar", "foo", "bar"]}
        mock_req.return_value = mockresponse
        mock_caid.return_value = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.cahandler._ca_policylink_id_lookup())
        self.assertIn(
            "ERROR:test_a2c:No policy link ID found for CA name: None",
            lcm.output,
        )
        self.assertIn("ERROR:test_a2c:No CAs found in issuer response.", lcm.output)
        self.assertFalse(mock_caid.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_bundle_build")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_id_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._csr_post")
    def test_069__cert_enroll(self, mock_post, mock_idget, mock_build):
        """test _cert_enroll()"""
        mock_post.return_value = "mock_post"
        mock_idget.return_value = "mock_idget"
        mock_build.return_value = ("error", "bundle", "raw")
        self.assertEqual(
            ("error", "bundle", "raw", "mock_idget"),
            self.cahandler._cert_enroll("cr", "policylink_id"),
        )
        self.assertTrue(mock_post.called)
        self.assertTrue(mock_idget.called)
        self.assertTrue(mock_build.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_bundle_build")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_id_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._csr_post")
    def test_070__cert_enroll(self, mock_post, mock_idget, mock_build):
        """test _cert_enroll()"""
        mock_post.return_value = "mock_post"
        mock_idget.return_value = None
        mock_build.return_value = ("error", "bundle", "raw")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("Certifcate_id lookup failed", None, None, None),
                self.cahandler._cert_enroll("cr", "policylink_id"),
            )
        self.assertIn(
            "ERROR:test_a2c:Certificate ID lookup failed for job: mock_post",
            lcm.output,
        )
        self.assertTrue(mock_post.called)
        self.assertTrue(mock_idget.called)
        self.assertFalse(mock_build.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_bundle_build")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_id_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._csr_post")
    def test_071__cert_enroll(self, mock_post, mock_idget, mock_build):
        """test _cert_enroll()"""
        mock_post.return_value = None
        mock_idget.return_value = "mock_idget"
        mock_build.return_value = ("error", "bundle", "raw")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("job_id lookup failed", None, None, None),
                self.cahandler._cert_enroll("cr", "policylink_id"),
            )
        self.assertIn(
            "ERROR:test_a2c:Job ID lookup failed during certificate enrollment.",
            lcm.output,
        )
        self.assertTrue(mock_post.called)
        self.assertFalse(mock_idget.called)
        self.assertFalse(mock_build.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._api_post")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.convert_string_to_byte")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.build_pem_file")
    def test_072__csr_post(self, mock_pem, mock_enc, mock_convert, mock_post):
        """test _csr_post()"""
        mock_pem.return_value = "mock_pem"
        mock_enc.return_value = "mock_enc"
        mock_convert.return_value = "mock_convert"
        mock_post.return_value = {"id": "id", "foo": "bar"}
        self.assertEqual("id", self.cahandler._csr_post("csr", "policylink_id"))
        self.assertTrue(mock_convert.called)
        self.assertTrue(mock_enc.called)
        self.assertTrue(mock_pem.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._api_post")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.convert_string_to_byte")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.build_pem_file")
    def test_073__csr_post(self, mock_pem, mock_enc, mock_convert, mock_post):
        """test _csr_post()"""
        mock_pem.return_value = "mock_pem"
        mock_enc.return_value = "mock_enc"
        mock_convert.return_value = "mock_convert"
        mock_post.return_value = {"foo": "bar"}
        self.cahandler.template_info_dic = {"id": "id"}
        self.assertFalse(self.cahandler._csr_post("csr", "policylink_id"))
        self.assertTrue(mock_convert.called)
        self.assertTrue(mock_enc.called)
        self.assertTrue(mock_pem.called)

    @patch("requests.get")
    def test_074__issuer_certid_get(self, mock_req):
        """test _issuer_certid_get()"""
        cert_dic = {"urls": {"issuer": "issuer"}}
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {"urls": {"certificate": "foo/v2/certificates/"}}
        mock_req.return_value = mockresponse
        self.assertEqual(("foo", True), self.cahandler._issuer_certid_get(cert_dic))

    @patch("requests.get")
    def test_075__issuer_certid_get(self, mock_req):
        """test _issuer_certid_get()"""
        cert_dic = {"urls": {"issuer": "issuer"}}
        self.cahandler.api_host = "api_host"
        mockresponse = Mock()
        mockresponse.json = lambda: {"urls": {"bar": "foo/v2/certificates/"}}
        mock_req.return_value = mockresponse
        self.assertEqual((None, False), self.cahandler._issuer_certid_get(cert_dic))

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._issuer_certid_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.build_pem_file")
    @patch("requests.get")
    def test_076__cert_bundle_build(self, mock_req, mock_pem, mock_certid):
        """test _cert_bundle_build()"""
        mock_pem.return_value = "mock_pem"
        mock_certid.return_value = ("id", False)
        mockresponse = Mock()
        mockresponse.json = lambda: {"der": "der"}
        mock_req.return_value = mockresponse
        self.assertEqual(
            (None, "mock_pem", "der"), self.cahandler._cert_bundle_build("cert_id")
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._issuer_certid_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.build_pem_file")
    @patch("requests.get")
    def test_077__cert_bundle_build(self, mock_req, mock_pem, mock_certid):
        """test _cert_bundle_build()"""
        mock_pem.side_effect = ["mock_pem1", "mock_pem2"]
        mock_certid.side_effect = [("id1", True), ("id2", False)]
        mockresponse = Mock()
        mockresponse.json = lambda: {"der": "der"}
        mock_req.return_value = mockresponse
        self.assertEqual(
            (None, "mock_pem2", "der"), self.cahandler._cert_bundle_build("cert_id")
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._issuer_certid_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.build_pem_file")
    @patch("requests.get")
    def test_078__cert_bundle_build(self, mock_req, mock_pem, mock_certid):
        """test _cert_bundle_build()"""
        mock_pem.return_value = ""
        mock_certid.return_value = ("id", False)
        mockresponse = Mock()
        mockresponse.json = lambda: {"der": "der"}
        mock_req.return_value = mockresponse
        self.assertEqual(
            (None, None, "der"), self.cahandler._cert_bundle_build("cert_id")
        )

    @patch("time.sleep")
    @patch("requests.get")
    def test_079__cert_id_get(self, mock_req, mock_sleep):
        """test _cert_id_get()"""
        mockresponse = Mock()
        mockresponse.json = lambda: {
            "status": "done",
            "entities": [{"ref": "certificate", "url": "foo/v2/certificates/"}],
        }
        mock_req.return_value = mockresponse
        self.assertEqual("foo", self.cahandler._cert_id_get(10))

    @patch("time.sleep")
    @patch("requests.get")
    def test_080__cert_id_get(self, mock_req, mock_sleep):
        """test _cert_id_get()"""
        mockresponse = Mock()
        mockresponse.json = lambda: {
            "status": "done",
            "entities": [{"foo": "certificate", "url": "foo/v2/certificates/"}],
        }
        mock_req.return_value = mockresponse
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.cahandler._cert_id_get(10))
        self.assertIn(
            "ERROR:test_a2c:Job completed but certificate reference is missing or malformed: {'status': 'done', 'entities': [{'foo': 'certificate', 'url': 'foo/v2/certificates/'}]}",
            lcm.output,
        )

    @patch("time.sleep")
    @patch("requests.get")
    def test_081__cert_id_get(self, mock_req, mock_sleep):
        """test _cert_id_get()"""
        mockresponse1 = Mock()
        mockresponse1.json = lambda: {
            "status": "note",
            "entities": [{"ref": "certificate", "url": "foo1/v2/certificates/"}],
        }
        mockresponse2 = Mock()
        mockresponse2.json = lambda: {
            "status": "done",
            "entities": [{"ref": "certificate", "url": "foo2/v2/certificates/"}],
        }
        mock_req.side_effect = [mockresponse1, mockresponse2]
        with self.assertLogs("test_a2c", level="DEBUG") as lcm:
            self.assertEqual("foo2", self.cahandler._cert_id_get(10))
        self.assertIn(
            "DEBUG:test_a2c:CAhandler._cert_id_get() waiting for job to complete. Attempt: 0 status: note",
            lcm.output,
        )

    @patch("requests.get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.cert_serial_get")
    def test_082__certid_get_from_serial(self, mock_serial, mock_req):
        """_certid_get_from_serial()"""
        mock_serial.return_value = "mock_serial"
        mockresponse = Mock()
        mockresponse.json = lambda: {"items": [{"id": "id1"}, {"id": "id2"}]}
        mock_req.return_value = mockresponse
        self.assertEqual("id1", self.cahandler._certid_get_from_serial("cert_raw"))

    @patch("requests.get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.cert_serial_get")
    def test_083__certid_get_from_serial(self, mock_serial, mock_req):
        """_certid_get_from_serial()"""
        mock_serial.return_value = "mock_serial"
        mockresponse = Mock()
        mockresponse.json = lambda: {"items": [{"di": "id1"}, {"id": "id2"}]}
        mock_req.return_value = mockresponse
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(None, self.cahandler._certid_get_from_serial("cert_raw"))
        self.assertIn(
            "ERROR:test_a2c:Failed to retrieve certificate by serial: mock_serial",
            lcm.output,
        )

    @patch("requests.get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.cert_serial_get")
    def test_084__certid_get_from_serial(self, mock_serial, mock_req):
        """_certid_get_from_serial()"""
        mock_serial.return_value = "mock_serial"
        mock_req.side_effect = Exception("mock_req")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(None, self.cahandler._certid_get_from_serial("cert_raw"))
        self.assertIn(
            "ERROR:test_a2c:API request to fetch certificates got aborted with err: mock_req",
            lcm.output,
        )
        self.assertIn(
            "ERROR:test_a2c:Failed to retrieve certificate by serial: mock_serial",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._certid_get_from_serial")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.header_info_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_085__cert_id_lookup(self, mock_enc, mock_info, mock_serial):
        """test _cert_id_lookup()"""
        mock_enc.return_value = "mock_enc"
        mock_info.return_value = [{"poll_identifier": "poll_identifier"}]
        mock_serial.return_value = "mock_serial"
        self.assertEqual("poll_identifier", self.cahandler._cert_id_lookup("cert_raw"))
        self.assertFalse(mock_serial.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._certid_get_from_serial")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.header_info_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_086__cert_id_lookup(self, mock_enc, mock_info, mock_serial):
        """test _cert_id_lookup()"""
        mock_enc.return_value = "mock_enc"
        mock_info.return_value = [{"poll_identifier": None}]
        mock_serial.return_value = "mock_serial"
        self.assertEqual("mock_serial", self.cahandler._cert_id_lookup("cert_raw"))
        self.assertTrue(mock_serial.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._certid_get_from_serial")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.header_info_get")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_087__cert_id_lookup(self, mock_enc, mock_info, mock_serial):
        """test _cert_id_lookup()"""
        mock_enc.return_value = "mock_enc"
        mock_info.return_value = [{"foo": "bar"}]
        mock_serial.return_value = "mock_serial"
        self.assertEqual("mock_serial", self.cahandler._cert_id_lookup("cert_raw"))
        self.assertTrue(mock_serial.called)

    @patch("time.sleep")
    @patch("requests.get")
    def test_088__revocation_status_poll(self, mock_req, mock_sleep):
        """test _revocation_status_poll()"""
        mockresponse = Mock()
        mockresponse.json = lambda: {"status": "done"}
        mock_req.return_value = mockresponse
        err_dic = {"serverinternal": "serverinternal"}
        self.assertEqual(
            (200, None, None),
            self.cahandler._revocation_status_poll("cert_id", err_dic),
        )

    @patch("time.sleep")
    @patch("requests.get")
    def test_089__revocation_status_poll(self, mock_req, mock_sleep):
        """test _revocation_status_poll()"""
        mockresponse = Mock()
        mockresponse.json = lambda: {"status": "failed"}
        mock_req.return_value = mockresponse
        err_dic = {"serverinternal": "serverinternal"}
        self.assertEqual(
            (500, "serverinternal", "Revocation operation failed: error from API"),
            self.cahandler._revocation_status_poll("cert_id", err_dic),
        )

    @patch("time.sleep")
    @patch("requests.get")
    def test_090__revocation_status_poll(self, mock_req, mock_sleep):
        """test _revocation_status_poll()"""
        mockresponse1 = Mock()
        mockresponse1.json = lambda: {"status": "pending"}
        mockresponse2 = Mock()
        mockresponse2.json = lambda: {"status": "done"}
        mock_req.side_effect = [mockresponse2, mockresponse2]
        err_dic = {"serverinternal": "serverinternal"}
        self.assertEqual(
            (200, None, None),
            self.cahandler._revocation_status_poll("cert_id", err_dic),
        )

    @patch("time.sleep")
    @patch("requests.get")
    def test_091__revocation_status_poll(self, mock_req, mock_sleep):
        """test _revocation_status_poll()"""
        mockresponse1 = Mock()
        mockresponse1.json = lambda: {"status": "pending"}
        mockresponse2 = Mock()
        mockresponse2.json = lambda: {"status": "failed"}
        mock_req.side_effect = [mockresponse2, mockresponse2]
        err_dic = {"serverinternal": "serverinternal"}
        self.assertEqual(
            (500, "serverinternal", "Revocation operation failed: error from API"),
            self.cahandler._revocation_status_poll("cert_id", err_dic),
        )

    @patch("time.sleep")
    @patch("requests.get")
    def test_092__revocation_status_poll(self, mock_req, mock_sleep):
        """test _revocation_status_poll()"""
        mockresponse = Mock()
        mockresponse.json = lambda: {"status": "pending"}
        mock_req.return_value = mockresponse
        err_dic = {"serverinternal": "serverinternal"}
        self.assertEqual(
            (500, "serverinternal", "Revocation operation failed: Timeout"),
            self.cahandler._revocation_status_poll("cert_id", err_dic),
        )

    @patch("acme2certifier.cahandlers.nclm_ca_handler.enrollment_config_log")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_enroll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_policylink_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_093_enroll(
        self, mock_recode, mock_policy, mock_template, mock_enroll, mock_ecl
    ):
        """test enroll"""
        mock_recode.return_value = "csr"
        mock_policy.return_value = "policylink_id"
        mock_template.return_value = "template_id"
        mock_enroll.return_value = ("error", "bundle", "raw", "cert_id")
        self.cahandler.template_info_dic = {"name": "name", "id": None}
        self.cahandler.container_info_dic = {"name": "name", "id": "id"}
        self.assertEqual(
            ("error", "bundle", "raw", "cert_id"), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_policy.called)
        self.assertTrue(mock_template.called)
        self.assertTrue(mock_enroll.called)
        self.assertFalse(mock_ecl.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.enrollment_config_log")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_enroll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_policylink_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_094_enroll(
        self, mock_recode, mock_policy, mock_template, mock_enroll, mock_ecl
    ):
        """test enroll"""
        mock_recode.return_value = "csr"
        mock_policy.return_value = "policylink_id"
        mock_template.return_value = "template_id"
        mock_enroll.return_value = ("error", "bundle", "raw", "cert_id")
        self.cahandler.enrollment_config_log = True
        self.cahandler.template_info_dic = {"name": "name", "id": None}
        self.cahandler.container_info_dic = {"name": "name", "id": "id"}
        self.assertEqual(
            ("error", "bundle", "raw", "cert_id"), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_policy.called)
        self.assertTrue(mock_template.called)
        self.assertTrue(mock_enroll.called)
        self.assertTrue(mock_ecl.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_enroll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_policylink_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_095_enroll(self, mock_recode, mock_policy, mock_template, mock_enroll):
        """test enroll"""
        mock_recode.return_value = "csr"
        mock_policy.return_value = "policylink_id"
        mock_template.return_value = "template_id"
        mock_enroll.return_value = ("error", "bundle", "raw", "cert_id")
        self.cahandler.template_info_dic = {"name": "name", "id": None}
        self.cahandler.container_info_dic = {"name": "name", "id": None}
        self.assertEqual(
            (
                'ID lookup for container"name" failed.',
                None,
                None,
                None,
            ),
            self.cahandler.enroll("csr"),
        )
        self.assertTrue(mock_recode.called)
        self.assertFalse(mock_policy.called)
        self.assertFalse(mock_template.called)
        self.assertFalse(mock_enroll.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_enroll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_policylink_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_096_enroll(self, mock_recode, mock_policy, mock_template, mock_enroll):
        """test enroll"""
        mock_recode.return_value = "csr"
        mock_policy.return_value = None
        mock_template.return_value = "template_id"
        mock_enroll.return_value = ("error", "bundle", "raw", "cert_id")
        self.cahandler.template_info_dic = {"name": "name", "id": None}
        self.cahandler.container_info_dic = {"name": "name", "id": "id"}
        self.assertEqual(
            ("Enrollment aborted. ca: None, tsg_id: id", None, None, None),
            self.cahandler.enroll("csr"),
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_policy.called)
        self.assertFalse(mock_template.called)
        self.assertFalse(mock_enroll.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_enroll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_policylink_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_097_enroll(self, mock_recode, mock_policy, mock_template, mock_enroll):
        """test enroll"""
        mock_recode.return_value = "csr"
        mock_policy.return_value = None
        mock_template.return_value = "template_id"
        mock_enroll.return_value = ("error", "bundle", "raw", "cert_id")
        self.cahandler.template_info_dic = {"name": "name", "id": None}
        self.cahandler.container_info_dic = {"name": "name", "id": "id"}
        self.cahandler.error = "error"
        self.assertEqual(("error", None, None, None), self.cahandler.enroll("csr"))
        self.assertTrue(mock_recode.called)
        self.assertFalse(mock_policy.called)
        self.assertFalse(mock_template.called)
        self.assertFalse(mock_enroll.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.eab_profile_header_info_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_enroll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._template_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._ca_policylink_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.b64_url_recode")
    def test_098_enroll(
        self, mock_recode, mock_policy, mock_template, mock_enroll, mock_eab
    ):
        """test enroll"""
        mock_recode.return_value = "csr"
        mock_policy.return_value = "policylink_id"
        mock_template.return_value = "template_id"
        mock_enroll.return_value = ("error", "bundle", "raw", "cert_id")
        mock_eab.return_value = "eab"
        self.cahandler.template_info_dic = {"name": "name", "id": None}
        self.cahandler.container_info_dic = {"name": "name", "id": "id"}
        self.assertEqual(("eab", None, None, None), self.cahandler.enroll("csr"))
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_policy.called)
        self.assertTrue(mock_template.called)
        self.assertFalse(mock_enroll.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._revocation_status_poll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._api_post")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.error_dic_get")
    def test_099_revoke(self, mock_err, mock_idl, mock_post, mock_poll):
        """test revoke"""
        mock_err.return_value = {"foo": "bar", "serverinternal": "serverinternal"}
        mock_idl.return_value = "cert_id"
        mock_post.return_value = {"urls": {"job": "foo/v2/jobs/"}}
        mock_poll.return_value = (200, "message", "detail")
        self.assertEqual((200, "message", "detail"), self.cahandler.revoke("cert_raw"))
        self.assertTrue(mock_err.called)
        self.assertTrue(mock_idl.called)
        self.assertTrue(mock_post.called)
        self.assertTrue(mock_poll.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.eab_profile_revocation_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._revocation_status_poll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._api_post")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.error_dic_get")
    def test_100_revoke(self, mock_err, mock_idl, mock_post, mock_poll, mock_eab):
        """test revoke"""
        mock_err.return_value = {"foo": "bar", "serverinternal": "serverinternal"}
        mock_idl.return_value = "cert_id"
        mock_post.return_value = {"urls": {"foo": "foo"}}
        mock_poll.return_value = (200, "message", "detail")
        self.assertEqual(
            (500, "serverinternal", "Revocation operation failed"),
            self.cahandler.revoke("cert_raw"),
        )
        self.assertTrue(mock_err.called)
        self.assertTrue(mock_idl.called)
        self.assertTrue(mock_post.called)
        self.assertFalse(mock_poll.called)
        self.assertFalse(mock_eab.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.eab_profile_revocation_check")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._revocation_status_poll")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._api_post")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._cert_id_lookup")
    @patch("acme2certifier.cahandlers.nclm_ca_handler.error_dic_get")
    def test_101_revoke(self, mock_err, mock_idl, mock_post, mock_poll, mock_eab):
        """test revoke"""
        mock_err.return_value = {"foo": "bar", "serverinternal": "serverinternal"}
        mock_idl.return_value = "cert_id"
        mock_post.return_value = {"urls": {"foo": "foo"}}
        mock_poll.return_value = (200, "message", "detail")
        self.cahandler.eab_profiling = True
        self.assertEqual(
            (500, "serverinternal", "Revocation operation failed"),
            self.cahandler.revoke("cert_raw"),
        )
        self.assertTrue(mock_err.called)
        self.assertTrue(mock_idl.called)
        self.assertTrue(mock_post.called)
        self.assertFalse(mock_poll.called)
        self.assertTrue(mock_eab.called)

    @patch("acme2certifier.cahandlers.nclm_ca_handler.CAhandler._config_check")
    def test_102_handler_check(self, mock_handler_check):
        """test handler_check"""
        self.cahandler.error = "mock_handler_check"
        self.assertEqual("mock_handler_check", self.cahandler.handler_check())


if __name__ == "__main__":

    if os.path.exists("acme_test.db"):
        os.remove("acme_test.db")
    unittest.main()



# -*- coding: utf-8 -*-
"""Renewalinfo class: ACME renewal info handler with separated config and repository helpers."""

from __future__ import print_function
from typing import Dict
from dataclasses import dataclass
from acme2certifier.acme_srv.db_handler import DBstore
from acme2certifier.acme_srv.message import Message
from acme2certifier.acme_srv.helper import (
    string_sanitize,
    certid_hex_get,
    uts_to_date_utc,
    error_dic_get,
    load_config,
    ca_handler_load,
    uts_now,
    cert_serial_get,
    cert_aki_get,
    b64_url_recode,
    b64_decode,
)


@dataclass
class RenewalinfoConfig:
    """configuration dataclass for Renewalinfo handler"""

    renewal_force: bool = False
    renewalthreshold_pctg: float = 85.0
    retry_after_timeout: int = 86400
    renewalinfo_lookup: bool = False


class RenewalinfoRepository:
    """Renewalinfo repository helper with database access methods."""

    def __init__(self, dbstore, logger):
        self.dbstore = dbstore
        self.logger = logger

    def get_certificate_by_certid(self, certid_hex):
        """Retrieve certificate by certid from database."""
        self.logger.debug(
            "RenewalinfoRepository.get_certificate_by_certid(%s)", certid_hex
        )
        try:
            return self.dbstore.certificate_lookup(
                "renewal_info",
                certid_hex,
                (
                    "id",
                    "name",
                    "cert",
                    "cert_raw",
                    "expire_uts",
                    "issue_uts",
                    "created_at",
                ),
            )
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to look up certificate for renewal info (draft01): %s",
                err_,
            )
            return None

    def get_certificates_by_serial(self, serial):
        """Retrieve certificates by serial from database."""
        self.logger.debug(
            "RenewalinfoRepository.get_certificates_by_serial(%s)", serial
        )
        try:
            return self.dbstore.certificates_search(
                "serial",
                serial,
                operant="is",
                vlist=[
                    "id",
                    "name",
                    "cert",
                    "cert_raw",
                    "expire_uts",
                    "issue_uts",
                    "aki",
                    "created_at",
                ],
            )
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to look up certificate for renewal info (draft02): %s",
                err_,
            )
            return []

    def add_certificate(self, data_dic):
        """Add or update certificate in database."""
        self.logger.debug("RenewalinfoRepository.add_certificate()")
        return self.dbstore.certificate_add(data_dic)

    def get_housekeeping_param(self, name):
        """Retrieve housekeeping parameter by name from database."""
        self.logger.debug("RenewalinfoRepository.get_housekeeping_param(%s)", name)
        return self.dbstore.hkparameter_get(name)

    def add_housekeeping_param(self, param):
        """Add or update housekeeping parameter in database."""
        self.logger.debug("RenewalinfoRepository.add_housekeeping_param()")
        return self.dbstore.hkparameter_add(param)


class Renewalinfo(object):
    """Renewalinfo handler with business logic, config, and repository helpers."""

    def __init__(
        self, debug: bool = False, srv_name: str = None, logger: object = None
    ):
        self.debug = debug
        self.logger = logger
        self.server_name = srv_name
        self.path_dic = {"renewalinfo": "/acme/renewal-info/"}
        self.dbstore = DBstore(self.debug, self.logger)
        self.message = Message(self.debug, self.server_name, self.logger)
        self.err_msg_dic = error_dic_get(self.logger)
        self.config = RenewalinfoConfig()
        self.repository = RenewalinfoRepository(self.dbstore, self.logger)
        self.cahandler = None

    def _load_configuration(self):
        """Load renewalinfo configuration from file (harmonized approach)"""
        self.logger.debug("Renewalinfo._load_configuration()")

        config_dic = load_config()

        if "Renewalinfo" in config_dic:
            try:
                self.config.renewal_force = config_dic.getboolean(
                    "Renewalinfo", "renewal_force", fallback=False
                )
            except Exception as err_:
                self.logger.error("renewal_force parsing error: %s", err_)
                self.config.renewal_force = False
            try:
                self.config.renewalthreshold_pctg = float(
                    config_dic.get(
                        "Renewalinfo", "renewalthreshold_pctg", fallback=85.0
                    )
                )
            except Exception as err_:
                self.logger.error("renewalthreshold_pctg parsing error: %s", err_)
                self.config.renewalthreshold_pctg = 85.0
            try:
                self.config.retry_after_timeout = int(
                    config_dic.get("Renewalinfo", "retry_after_timeout", fallback=86400)
                )
            except Exception as err_:
                self.logger.error("retry_after_timeout parsing error: %s", err_)
                self.config.retry_after_timeout = 86400

        self._load_ca_handler(config_dic)
        self._parse_cahandler_section(config_dic)

        self.logger.debug("Renewalinfo._load_configuration() ended.")

    def _parse_cahandler_section(self, config_dic: object) -> None:
        """Parse the [CAHandler] section for ACME URL and profile sync settings."""
        self.logger.debug("Directory._parse_cahandler_section()")
        if "CAhandler" in config_dic:
            cfg_dic = dict(config_dic["CAhandler"])
            self.config.acme_url = cfg_dic.get("acme_url", None)

            try:
                self.config.renewalinfo_lookup = config_dic.getboolean(
                    "CAhandler", "renewalinfo_lookup", fallback=False
                )
            except Exception as err_:
                self.logger.error("renewalinfo_lookup parsing error: %s", err_)
                self.config.renewalinfo_lookup = False

        if self.config.renewalinfo_lookup and not self.config.acme_url:
            self.logger.error("CAhandler section incomplete for renewalinfo lookup")
            self.config.renewalinfo_lookup = False

        self.logger.debug("Directory._parse_cahandler_section() ended")

    def _load_ca_handler(self, config_dic: object) -> None:
        """Load the CA handler module as configured."""
        ca_handler_module = ca_handler_load(self.logger, config_dic)
        if ca_handler_module:
            self.cahandler = ca_handler_module.CAhandler
        else:
            self.logger.critical("No ca_handler loaded")

    def __enter__(self):
        self._load_configuration()
        return self

    def __exit__(self, *args):
        pass

    # --- Business logic methods ---

    def _lookup_certificate_by_renewalinfo(
        self, renewalinfo_string: str
    ) -> Dict[str, str]:
        self.logger.debug(
            "Renewalinfo._lookup_certificate_by_renewalinfo(%s)", renewalinfo_string
        )
        if "." in renewalinfo_string:
            serial, aki = self._extract_serial_and_aki_from_string(renewalinfo_string)
            cert_dic = self._lookup_certificate_by_serial_and_aki(serial, aki)
        else:
            _mda, certid_hex = certid_hex_get(self.logger, renewalinfo_string)
            cert_dic = self._lookup_certificate_by_certid(certid_hex)
        self.logger.debug(
            "Renewalinfo._lookup_certificate_by_renewalinfo(%s) - ended with: %s",
            renewalinfo_string,
            bool(cert_dic),
        )
        return cert_dic

    def _update_certificate_table_with_serial_and_aki(self):
        self.logger.debug("Renewalinfo._update_certificate_table_with_serial_and_aki()")
        try:
            certificate_list = self.dbstore.certificates_search(
                "serial",
                None,
                operant="is",
                vlist=["id", "name", "cert", "cert_raw", "serial", "aki"],
            )
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to retrieve certificate list for renewal info update: %s",
                err_,
            )
            certificate_list = []
        update_cnt = 0
        for cert in certificate_list:
            if (
                "cert_raw" in cert
                and cert["cert_raw"]
                and "name" in cert
                and "cert" in cert
            ):
                serial = cert_serial_get(self.logger, cert["cert_raw"], hexformat=True)
                aki = cert_aki_get(self.logger, cert["cert_raw"])
                data_dic = {
                    "serial": serial,
                    "aki": aki,
                    "name": cert["name"],
                    "cert_raw": cert["cert_raw"],
                    "cert": cert["cert"],
                }
                self.repository.add_certificate(data_dic)
                update_cnt += 1
        self.logger.debug(
            "Renewalinfo._update_certificate_table_with_serial_and_aki(%s) - done",
            update_cnt,
        )

    def _lookup_certificate_by_certid(self, certid_hex: str) -> Dict[str, str]:
        self.logger.debug("Renewalinfo._lookup_certificate_by_certid()")
        return self.repository.get_certificate_by_certid(certid_hex)

    def _lookup_certificate_by_serial_and_aki(
        self, serial: str, aki: str
    ) -> Dict[str, str]:
        self.logger.debug("Renewalinfo._lookup_certificate_by_serial_and_aki()")
        cert_dic = {}
        cert_list = self.repository.get_certificates_by_serial(serial)
        if not cert_list and serial and serial.startswith("0"):
            cert_list = self.repository.get_certificates_by_serial(serial.lstrip("0"))
        for cert in cert_list:
            if cert.get("aki") == aki:
                cert_dic = cert
                break
        self.logger.debug(
            "Renewalinfo._lookup_certificate_by_serial_and_aki() ended with: %s",
            bool(cert_dic),
        )
        return cert_dic

    def _generate_renewalinfo_window(self, cert_dic: Dict[str, str]) -> Dict[str, str]:
        self.logger.debug("Renewalinfo._generate_renewalinfo_window()")
        if "expire_uts" in cert_dic and cert_dic["expire_uts"]:
            if "issue_uts" not in cert_dic or not cert_dic["issue_uts"]:
                cert_dic["issue_uts"] = uts_now()
            if self.config.renewal_force:
                self.logger.debug("Renewalinfo.get() - force renewal")
                cert_dic["expire_uts"] = uts_now() + 86400
                start_uts = int(cert_dic["expire_uts"] - (365 * 86400))
            else:
                start_uts = (
                    int(
                        (cert_dic["expire_uts"] - cert_dic["issue_uts"])
                        * self.config.renewalthreshold_pctg
                        / 100
                    )
                    + cert_dic["issue_uts"]
                )
            renewalinfo_dic = {
                "suggestedWindow": {
                    "start": uts_to_date_utc(start_uts),
                    "end": uts_to_date_utc(cert_dic["expire_uts"]),
                }
            }
        else:
            renewalinfo_dic = {}
        self.logger.debug("Renewalinfo._generate_renewalinfo_window() ended")
        return renewalinfo_dic

    def _get_renewalinfo_data(self, renewalinfo_string: str) -> Dict[str, str]:
        self.logger.debug("Renewalinfo._get_renewalinfo_data()")
        cert_dic = self._lookup_certificate_by_renewalinfo(renewalinfo_string)
        renewalinfo_dic = self._generate_renewalinfo_window(cert_dic)
        self.logger.debug(
            "Renewalinfo._get_renewalinfo_data() ended with: %s", renewalinfo_dic
        )
        return renewalinfo_dic

    def _parse_renewalinfo_string_from_url(self, url: str) -> str:
        self.logger.debug("Renewalinfo._parse_renewalinfo_string_from_url()")
        url = url.replace(
            f'{self.server_name}{self.path_dic["renewalinfo"].rstrip("/")}', ""
        )
        url = url.lstrip("/")
        renewalinfo_string = string_sanitize(self.logger, url)
        self.logger.debug(
            "Renewalinfo._parse_renewalinfo_string_from_url() - renewalinfo_string: %s",
            renewalinfo_string,
        )
        return renewalinfo_string

    def _extract_serial_and_aki_from_string(
        self, renewalinfo_string: str
    ) -> (str, str):
        self.logger.debug("Renewalinfo._extract_serial_and_aki_from_string()")
        renewalinfo_list = renewalinfo_string.split(".")
        if len(renewalinfo_list) == 2:
            serial = b64_decode(
                self.logger, b64_url_recode(self.logger, renewalinfo_list[1])
            ).hex()
            aki = b64_decode(
                self.logger, b64_url_recode(self.logger, renewalinfo_list[0])
            ).hex()
        else:
            serial = None
            aki = None
        self.logger.debug(
            "Renewalinfo._extract_serial_and_aki_from_string() - serial: %s, aki: %s",
            serial,
            aki,
        )
        return (serial, aki)

    def get(self, url: str) -> Dict[str, str]:
        """Get renewal information (backwards compatible public method)"""
        self.logger.debug("Renewalinfo.get()")

        renewalinfo_string = self._parse_renewalinfo_string_from_url(url)
        if self.config.renewalinfo_lookup and hasattr(
            self.cahandler, "lookup_renewalinfo"
        ):
            with self.cahandler(None, self.logger) as ca_handler:
                # get renewal info from CA handler
                rc_code, renewalinfo_dic = ca_handler.lookup_renewalinfo(
                    self.config.acme_url, renewalinfo_string
                )
        else:
            # ensure serial and aki fields are populated in certificate table
            if not self.repository.get_housekeeping_param("cert_aki_serial_update"):
                self._update_certificate_table_with_serial_and_aki()
                self.logger.debug("Renewalinfo.get() - update housekeeping")
                self.repository.add_housekeeping_param(
                    {"name": "cert_aki_serial_update", "value": True}
                )

            try:
                renewalinfo_dic = self._get_renewalinfo_data(renewalinfo_string)
                rc_code = 200 if renewalinfo_dic else 404
            except Exception as err_:
                self.logger.error("Error when getting renewal information: %s", err_)
                renewalinfo_dic = {}
                rc_code = 400
        response_dic = {"code": rc_code}
        if renewalinfo_dic:
            response_dic["data"] = renewalinfo_dic
            response_dic["header"] = {
                "Retry-After": f"{self.config.retry_after_timeout}"
            }
        else:
            response_dic["data"] = self.err_msg_dic["malformed"]
        return response_dic

    def update(self, content: str) -> Dict[str, str]:
        """Update renewal info (backwards compatible public method)"""
        self.logger.debug("Renewalinfo.update()")
        (
            code,
            _message,
            _detail,
            _protected,
            payload,
            _account_name,
        ) = self.message.check(content)
        response_dic = {}
        if code == 200 and "certid" in payload and "replaced" in payload:
            cert_dic = self._lookup_certificate_by_renewalinfo(payload["certid"])
            if cert_dic and payload["replaced"]:
                cert_dic["replaced"] = True
                cert_id = self.repository.add_certificate(cert_dic)
                response_dic["code"] = 200 if cert_id else 400
            else:
                response_dic["code"] = 400
        else:
            response_dic["code"] = 400
        return response_dic



# -*- coding: utf-8 -*-
"""ca handler for Insta Certifier via REST-API class"""

from __future__ import print_function
import textwrap
import math
import time
import json
import os
from typing import List, Tuple, Dict
import requests
from requests.auth import HTTPBasicAuth

# pylint: disable=e0401
from acme2certifier.acme_srv.helper import (
    b64_decode,
    b64_encode,
    cert_pem2der,
    cert_serial_get,
    config_eab_profile_load,
    config_enroll_config_log_load,
    config_headerinfo_load,
    config_profile_load,
    eab_profile_header_info_check,
    eab_profile_revocation_check,
    enrollment_config_log,
    error_dic_get,
    handler_config_check,
    load_config,
    parse_url,
    proxy_check,
    uts_now,
    uts_to_date_utc,
)


class CAhandler(object):
    """CA  handler"""

    def __init__(self, debug: bool = False, logger: object = None):
        self.debug = debug
        self.logger = logger
        self.request_timeout = 20
        self.api_host = None
        self.api_user = None
        self.api_password = None
        self.ca_bundle = True
        self.ca_name = None
        self.auth = None
        self.polling_timeout = 60
        self.profile_id = None
        self.proxy = None
        self.header_info_field = False
        self.eab_handler = None
        self.eab_profiling = False
        self.enrollment_config_log = False
        self.enrollment_config_log_skip_list = []
        self.profiles = {}
        self.profile_mapping_field = "profile_id"

    def __enter__(self):
        """Makes ACMEHandler a Context Manager"""
        if not self.api_host:
            self._config_load()
            self._auth_set()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _auth_set(self):
        """set basic authentication header"""
        self.logger.debug("CAhandler._auth_set()")
        if self.api_user and self.api_password:
            self.auth = HTTPBasicAuth(self.api_user, self.api_password)
        else:
            self.logger.error(
                'Auth information incomplete. Either "api_user" or "api_password" parameter is missing in config file'
            )
        self.logger.debug("CAhandler._auth_set() ended")

    def _api_poll(self, request_dic: Dict[str, str]) -> Tuple[str, str, str]:
        """poll request"""
        self.logger.debug("CAhandler._api_poll()")

        cert_bundle = None
        cert_raw = None

        if "certificate" in request_dic:
            # poll identifier for later storage
            cert_dic = requests.get(
                request_dic["certificate"],
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
            if "certificateBase64" in cert_dic:
                # this is a valid cert generate the bundle
                error = None
                cert_bundle = self._pem_cert_chain_generate(cert_dic)
                cert_raw = cert_dic["certificateBase64"]
            else:
                error = "certificateBase64 is missing in cert request response"
        else:
            error = "No certificate structure in request response"

        self.logger.debug("CAhandler._api_poll() ended")
        return (error, cert_bundle, cert_raw)

    def _api_post(self, url: str, data: Dict[str, str]) -> Dict[str, str]:
        """
        generic wrapper for an API post call
        args:
            url - API URL
            data - data to post
        returns:
            result of the post command
        """
        try:
            api_response = requests.post(
                url=url,
                json=data,
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        except Exception as err_:
            self.logger.error("API post() request returned an error: %s", err_)
            api_response = str(err_)

        return api_response

    def _ca_get(
        self, filter_key: str = None, filter_value: str = None
    ) -> Dict[str, str]:
        """get list of CAs"""
        self.logger.debug("_ca_get(%s:%s)", filter_key, filter_value)
        params = {}

        if filter_key:
            params["q"] = f"{filter_key}:{filter_value}"

        if self.api_host:
            try:
                api_response = requests.get(
                    self.api_host + "/v1/cas",
                    auth=self.auth,
                    params=params,
                    proxies=self.proxy,
                    verify=self.ca_bundle,
                    timeout=self.request_timeout,
                ).json()
            except Exception as err_:
                self.logger.error("API get() request returned error: %s", str(err_))
                api_response = {
                    "status": 500,
                    "message": str(err_),
                    "statusMessage": "Internal Server Error",
                }
        else:
            self.logger.error("api_host parameter is misisng in configuration")
            api_response = {}
        self.logger.debug("CAhandler._ca_get() ended with: %s", api_response)
        return api_response

    def _ca_get_properties(self, filter_key: str, filter_value: str) -> Dict[str, str]:
        """get properties for a single CAs"""
        self.logger.debug("_ca_get_properties(%s:%s)", filter_key, filter_value)
        ca_list = self._ca_get(filter_key, filter_value)
        ca_dic = {}
        if "status" in ca_list and "message" in ca_list:
            # we got an error from get_ca()
            ca_dic = ca_list
        elif "cas" in ca_list and ca_list["cas"]:
            for cas in ca_list["cas"]:
                if filter_key in cas and cas[filter_key] == filter_value:
                    ca_dic = cas
                    break
        if not ca_dic:
            ca_dic = {
                "status": 404,
                "message": "CA not found",
                "statusMessage": "Not Found",
            }
        self.logger.debug("CAhandler._ca_get_properties() ended with: %s", ca_dic)
        return ca_dic

    def _cert_get(self, csr: str) -> Dict[str, str]:
        """get certificate from CA"""
        self.logger.debug("CAhandler._cert_get(%s)", csr)
        ca_dic = self._ca_get_properties("name", self.ca_name)
        cert_dic = {}

        if self.enrollment_config_log:
            self.enrollment_config_log_skip_list.extend(["auth", "api_password"])
            enrollment_config_log(
                self.logger, self, self.enrollment_config_log_skip_list
            )

        if "href" in ca_dic:
            data = {"ca": ca_dic["href"], "pkcs10": csr}

            # set profileid if configured
            if self.profile_id:
                data["profileId"] = self.profile_id

            cert_dic = self._api_post(self.api_host + "/v1/requests", data)

        if not cert_dic:
            cert_dic = ca_dic

        self.logger.debug("CAhandler._cert_get() ended with: %s", cert_dic)
        return cert_dic

    def _cert_get_properties(self, serial: str, ca_link: str) -> Dict[str, str]:
        """get properties for a single cert"""
        self.logger.debug("_cert_get_properties(%s:%s)", serial, ca_link)

        params = {"q": f"issuer-id:{ca_link},serial-number:{serial}"}
        try:
            api_response = requests.get(
                self.api_host + "/v1/certificates",
                auth=self.auth,
                params=params,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        except Exception as err_:
            self.logger.error(
                "Could not get certificate properties. Error: %s", str(err_)
            )
            api_response = {
                "status": 500,
                "message": str(err_),
                "statusMessage": "Internal Server Error",
            }
        self.logger.debug("CAhandler._cert_get_properties() ended")
        return api_response

    def _certificate_revoke(
        self, serial: str, ca_dic: Dict[str, str], rev_reason: str, rev_date: str
    ) -> Tuple[int, str, str]:
        self.logger.debug("CAhandler._certificate_revoke()")

        code = None
        message = None
        detail = None

        # get error message
        err_dic = error_dic_get(self.logger)

        # get certificate information via rest by search for ca+ serial
        cert_dic = self._cert_get_properties(serial, ca_dic["href"])
        if "certificates" in cert_dic:
            if (
                len(cert_dic["certificates"]) > 0
                and "href" in cert_dic["certificates"][0]
            ):
                # revoke the cert
                data = {
                    "newStatus": "revoked",
                    "crlReason": rev_reason,
                    "invalidityDate": rev_date,
                }
                cert_dic = self._api_post(
                    cert_dic["certificates"][0]["href"] + "/status", data
                )
                if "status" in cert_dic:
                    code = 400
                    message = err_dic["alreadyrevoked"]
                    if "message" in cert_dic:
                        detail = cert_dic["message"]
                    else:
                        detail = "no details"
                else:
                    code = 200
                    message = None
                    detail = None
            else:
                code = 404
                message = err_dic["serverinternal"]
                detail = "Cert path could not be found"
        else:
            code = 404
            message = err_dic["serverinternal"]
            detail = "Cert could not be found"

        return (code, message, detail)

    def _config_user_load(self, config_dic: Dict[str, str]):
        """load username"""
        self.logger.debug("_config_user_load()")
        if (
            "api_user" in config_dic["CAhandler"]
            or "api_user_variable" in config_dic["CAhandler"]
        ):
            if "api_user_variable" in config_dic["CAhandler"]:
                try:
                    self.api_user = os.environ[
                        config_dic.get("CAhandler", "api_user_variable")
                    ]
                except Exception as err:
                    self.logger.error("Could not load user_variable:%s", err)
            if "api_user" in config_dic["CAhandler"]:
                if self.api_user:
                    self.logger.info("Overwrite api_user")
                self.api_user = config_dic.get(
                    "CAhandler", "api_user", fallback=self.api_user
                )
        else:
            self.logger.error(
                'Configuration incomplete: "api_user" parameter is missing in config file'
            )

        self.logger.debug("_config_user_load() ended")

    def _config_password_load(self, config_dic: Dict[str, str]):
        """load password"""
        self.logger.debug("_config_password_load()")

        if (
            "api_password" in config_dic["CAhandler"]
            or "api_password_variable" in config_dic["CAhandler"]
        ):
            if "api_password_variable" in config_dic["CAhandler"]:
                try:
                    self.api_password = os.environ[
                        config_dic.get("CAhandler", "api_password_variable")
                    ]
                except Exception as err:
                    self.logger.error(
                        "Could not load passphrase_variable:%s",
                        err,
                    )
            if "api_password" in config_dic["CAhandler"]:
                if self.api_password:
                    self.logger.info("Overwrite api_password_variable")
                self.api_password = config_dic.get("CAhandler", "api_password")
        else:
            self.logger.error(
                'Configuration incomplete: "api_password" parameter is missing in config file'
            )

        self.logger.debug("_config_password_load() ended")

    def _config_parameter_load(self, config_dic: Dict[str, str]):
        """load parameters"""
        self.logger.debug("_config_parameter_load()")

        if "ca_name" in config_dic["CAhandler"]:
            self.ca_name = config_dic.get("CAhandler", "ca_name", fallback=self.ca_name)
        else:
            self.logger.error(
                'Configuration incomplete: "ca_name" parameter is missing in config file'
            )

        try:
            self.polling_timeout = int(
                config_dic.get(
                    "CAhandler", "polling_timeout", fallback=self.polling_timeout
                )
            )
        except Exception:
            self.logger.warning(
                "Invalid value for polling_timeout in configuration. Using default: %s",
                self.polling_timeout,
            )

        try:
            self.request_timeout = int(
                config_dic.get(
                    "CAhandler", "request_timeout", fallback=self.request_timeout
                )
            )
        except Exception:
            self.logger.warning(
                "Invalid value for request_timeout in configuration. Using default: %s",
                self.request_timeout,
            )

        # load enrollment config log
        (
            self.enrollment_config_log,
            self.enrollment_config_log_skip_list,
        ) = config_enroll_config_log_load(self.logger, config_dic)

        # load profile_id
        self.profile_id = config_dic.get(
            "CAhandler", self.profile_mapping_field, fallback=None
        )

        # check if we get a ca bundle for verification
        if "ca_bundle" in config_dic["CAhandler"]:
            try:
                self.ca_bundle = config_dic.getboolean("CAhandler", "ca_bundle")
            except Exception:
                self.ca_bundle = config_dic.get(
                    "CAhandler", "ca_bundle", fallback=self.ca_bundle
                )

        self.logger.debug("_config_parameter_load() ended")

    def _config_proxy_load(self, config_dic: Dict[str, str]):
        """load parameters"""
        self.logger.debug("_config_proxy_load()")

        if "DEFAULT" in config_dic and "proxy_server_list" in config_dic["DEFAULT"]:
            try:
                proxy_list = json.loads(config_dic["DEFAULT"]["proxy_server_list"])
                url_dic = parse_url(self.logger, self.api_host)
                if "host" in url_dic:
                    fqdn, _port = url_dic["host"].split(":")
                    proxy_server = proxy_check(self.logger, fqdn, proxy_list)
                    self.proxy = {"http": proxy_server, "https": proxy_server}
            except Exception as err_:
                self.logger.warning(
                    "Failed to parse proxy_server_list from configuration: %s",
                    err_,
                )

        self.logger.debug("_config_proxy_load() ended")

    def _config_load(self):
        """ " load config from file"""
        # pylint: disable=R0912, R0915
        self.logger.debug("_config_load()")
        config_dic = load_config(self.logger, "CAhandler")
        if "CAhandler" in config_dic:
            if "api_host" in config_dic["CAhandler"]:
                self.api_host = config_dic.get(
                    "CAhandler", "api_host", fallback=self.api_host
                )
            else:
                self.logger.error(
                    'Configuration incomplete: "api_host" parameter is missing in config file'
                )

            # load user from config
            self._config_user_load(config_dic)
            # load password from config
            self._config_password_load(config_dic)
            # load parameters from config
            self._config_parameter_load(config_dic)
            # load profiling
            self.eab_profiling, self.eab_handler = config_eab_profile_load(
                self.logger, config_dic
            )

            # load profiles
            self.profiles = config_profile_load(self.logger, config_dic)

            # load header info
            self.header_info_field = config_headerinfo_load(self.logger, config_dic)

        # load proxy configuration
        self._config_proxy_load(config_dic)
        self.logger.debug("CAhandler._config_load() ended")

    def _csr_check(self, csr: str) -> str:
        """check csr"""
        self.logger.debug("CAhandler._csr_check()")

        # check for eab profiling and header_info
        error = eab_profile_header_info_check(
            self.logger, self, csr, self.profile_mapping_field
        )

        self.logger.debug("CAhandler._csr_check() ended with: %s", error)
        return error

    def _poll_cert_get(
        self, request_dic: Dict[str, str], poll_identifier: str, error: str
    ) -> Tuple[str, str, str, str, bool]:
        """get certificate via poll request"""
        self.logger.debug("CAhandler._poll_cert_get()")

        cert_bundle = None
        cert_raw = None
        break_loop = False
        # check response
        if "status" in request_dic:
            if request_dic["status"] == "accepted":

                if "certificate" in request_dic:
                    # poll identifier for later storage
                    cert_dic = requests.get(
                        request_dic["certificate"],
                        auth=self.auth,
                        verify=self.ca_bundle,
                        proxies=self.proxy,
                        timeout=self.request_timeout,
                    ).json()
                    # pylint: disable=R1723
                    if "certificateBase64" in cert_dic:
                        # this is a valid cert generate the bundle
                        error = None
                        cert_bundle = self._pem_cert_chain_generate(cert_dic)
                        cert_raw = cert_dic["certificateBase64"]
                        poll_identifier = None
                        break_loop = True
                    else:
                        error = "Request accepted but no certificateBase64 returned"
                else:
                    error = "Request accepted but no certificate returned"
            elif request_dic["status"] == "rejected":
                error = "Request rejected by operator"
                poll_identifier = None
                break_loop = True

        self.logger.debug("CAhandler._poll_cert_get() ended")
        return (error, cert_bundle, cert_raw, poll_identifier, break_loop)

    def _loop_poll(self, request_url: str) -> Tuple[str, str, str, str]:
        """poll request"""
        self.logger.debug("CAhandler._loop_poll(%s)", request_url)

        error = None
        cert_bundle = None
        cert_raw = None

        if request_url:
            # calculate iterations based on timeout
            poll_cnt = math.ceil(self.polling_timeout / 5)
            cnt = 1
            while cnt <= poll_cnt:
                cnt += 1
                request_dic = requests.get(
                    request_url,
                    auth=self.auth,
                    verify=self.ca_bundle,
                    proxies=self.proxy,
                    timeout=self.request_timeout,
                ).json()

                # check response
                (
                    error,
                    cert_bundle,
                    cert_raw,
                    poll_identifier,
                    break_loop,
                ) = self._poll_cert_get(request_dic, request_url, error)
                if break_loop:
                    break

                # sleep
                time.sleep(self.request_timeout)
        else:
            self.logger.warning("Error during polling loop: no request url specified")
            poll_identifier = request_url

        self.logger.debug("CAhandler._loop_poll() ended with error: %s", error)
        return (error, cert_bundle, cert_raw, poll_identifier)

    def _pem_list_cert_get(self, cert_dic: Dict[str, str]) -> Dict[str, str]:
        self.logger.debug("CAhandler._pem_list_cert_get()")
        if "issuer" in cert_dic:
            self.logger.debug("issuer found: %s", cert_dic["issuer"])
            ca_cert_dic = requests.get(
                cert_dic["issuer"],
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        else:
            self.logger.debug("issuer found: %s", cert_dic["issuerCa"])
            ca_cert_dic = requests.get(
                cert_dic["issuerCa"],
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()

        cert_dic = {}
        if "certificates" in ca_cert_dic:
            if "active" in ca_cert_dic["certificates"]:
                cert_dic = requests.get(
                    ca_cert_dic["certificates"]["active"],
                    auth=self.auth,
                    verify=self.ca_bundle,
                    proxies=self.proxy,
                    timeout=self.request_timeout,
                ).json()

        self.logger.debug("CAhandler._pem_list_cert_get() ended")
        return cert_dic

    def _pem_list_build(self, cert_dic: Dict[str, str]) -> List[str]:
        self.logger.debug("CAhandler._pem_list_build()")

        pem_list = []
        issuer_loop = True
        while issuer_loop:
            if "certificateBase64" in cert_dic:
                pem_list.append(cert_dic["certificateBase64"])
            else:
                # stop if there is no pem content in the json response
                issuer_loop = False  # lgtm [py/unused-local-variable]
                break
            if "issuer" in cert_dic or "issuerCa" in cert_dic:
                cert_dic = self._pem_list_cert_get(cert_dic)
            else:
                issuer_loop = False  # lgtm [py/unused-local-variable]
                break

        self.logger.debug("CAhandler._pem_list_build() ended")
        return pem_list

    def _pem_cert_chain_generate(self, cert_dic: str) -> str:
        """build certificate chain based"""
        self.logger.debug("CAhandler._pem_cert_chain_generate()")

        if cert_dic:
            pem_list = self._pem_list_build(cert_dic)
        else:
            pem_list = []

        if pem_list:
            pem_file = ""
            for cert in pem_list:
                pem_file = f"{pem_file}-----BEGIN CERTIFICATE-----\n{textwrap.fill(cert, 64)}\n-----END CERTIFICATE-----\n"
        else:
            pem_file = None

        self.logger.debug("CAhandler._pem_cert_chain_generate() ended")
        return pem_file

    def _request_poll(self, request_url: str) -> Tuple[str, str, str, str, bool]:
        """poll request"""
        self.logger.debug("CAhandler._request_poll(%s)", request_url)

        error = None
        cert_bundle = None
        cert_raw = None
        poll_identifier = request_url
        rejected = False

        try:
            request_dic = requests.get(
                request_url,
                auth=self.auth,
                verify=self.ca_bundle,
                proxies=self.proxy,
                timeout=self.request_timeout,
            ).json()
        except Exception as err:
            self.logger.error("Polling request returned an error: %s", err)
            request_dic = {}

        # check response
        if "status" in request_dic:
            if request_dic["status"] == "accepted":
                error, cert_bundle, cert_raw = self._api_poll(request_dic)
            elif request_dic["status"] == "rejected":
                error = "Request rejected by operator"
                rejected = True
            else:
                error = f'Unknown request status: {request_dic["status"]}'
        else:
            error = '"status" field not found in response.'

        self.logger.debug("CAhandler._request_poll() ended with error: %s", error)
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def _trigger_bundle_build(
        self, cert_raw: str, ca_dic: Dict[str, str]
    ) -> Tuple[str, str]:
        self.logger.debug("CAhandler._trigger_bundle_build()")
        error = None
        cert_bundle = None

        # get serial from pem file
        serial = cert_serial_get(self.logger, cert_raw)
        if serial:
            # get certificate information via rest by search for ca+ serial
            cert_list = self._cert_get_properties(serial, ca_dic["href"])
            # the first entry is the cert we are looking for
            if "certificates" in cert_list and len(cert_list["certificates"][0]) > 0:
                cert_dic = cert_list["certificates"][0]
                cert_bundle = self._pem_cert_chain_generate(cert_dic)
            else:
                error = "no certifcates found in rest query"
        else:
            error = "serial number lookup via rest failed"

        self.logger.debug("CAhandler._trigger_bundle_build() ended with:  %s", error)
        return (error, cert_bundle)

    def enroll(self, csr: str) -> Tuple[str, str, str, str]:
        """enroll certificate"""
        self.logger.debug("Certificate.enroll()")
        cert_bundle = None
        cert_raw = None
        poll_identifier = None

        # check CSR
        error = self._csr_check(csr)

        # enrollment starts here
        if not error:
            cert_dic = self._cert_get(csr)
        else:
            cert_dic = None

        if cert_dic:
            if "status" in cert_dic:
                # this is an error
                if "message" in cert_dic:
                    error = cert_dic["message"]
                else:
                    error = "unknown error"
            elif "certificateBase64" in cert_dic:
                # this is a valid cert generate the bundle
                cert_bundle = self._pem_cert_chain_generate(cert_dic)
                cert_raw = cert_dic["certificateBase64"]
            elif "href" in cert_dic:
                # request is pending
                error, cert_bundle, cert_raw, poll_identifier = self._loop_poll(
                    cert_dic["href"]
                )
            else:
                error = "no certificate information found"
        else:
            if not error:
                error = "internal error"

        self.logger.debug("Certificate.enroll() ended")
        return (error, cert_bundle, cert_raw, poll_identifier)

    def handler_check(self):
        """check if handler is ready"""
        self.logger.debug("CAhandler.check()")

        error = handler_config_check(
            self.logger, self, ["api_host", "api_user", "api_password", "ca_name"]
        )

        self.logger.debug("CAhandler.check() ended with %s", error)
        return error

    def poll(
        self, cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[str, str, str, str, str, bool]:
        """poll pending status of pending CSR and download certificates"""
        self.logger.debug("CAhandler.poll()")

        error = None
        cert_bundle = None
        cert_raw = None
        rejected = False

        if poll_identifier:
            (
                error,
                cert_bundle,
                cert_raw,
                poll_identifier,
                rejected,
            ) = self._request_poll(poll_identifier)
        else:
            self.logger.debug(
                "skipping cert: %s as there is no poll_identifier", cert_name
            )

        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def revoke(
        self,
        cert: str,
        rev_reason: str = "unspecified",
        rev_date: str = uts_to_date_utc(uts_now()),
    ) -> Tuple[int, str, str]:
        """revoke certificate"""
        self.logger.debug("CAhandler.revoke(%s: %s)", rev_reason, rev_date)

        # get error message
        err_dic = error_dic_get(self.logger)

        # modify handler configuration in case of eab profiling
        if self.eab_profiling:
            eab_profile_revocation_check(self.logger, self, cert)

        # lookup REST-PATH of issuing CA
        ca_dic = self._ca_get_properties("name", self.ca_name)
        if "href" in ca_dic:
            # get serial from pem file
            serial = cert_serial_get(self.logger, cert)
            if serial:
                code, message, detail = self._certificate_revoke(
                    serial, ca_dic, rev_reason, rev_date
                )
            else:
                code = 404
                message = err_dic["serverinternal"]
                detail = "failed to get serial number from cert"
        else:
            code = 404
            message = err_dic["serverinternal"]
            detail = "CA could not be found"

        return (code, message, detail)

    def trigger(self, payload: str) -> Tuple[str, str, str]:
        """process trigger message and return certificate"""
        self.logger.debug("CAhandler.trigger()")

        error = None
        cert_bundle = None
        cert_raw = None

        if payload:
            # decode payload
            cert = b64_decode(self.logger, payload)
            try:
                # cert is a base64 encoded pem object
                cert_raw = b64_encode(self.logger, cert_pem2der(cert))
            except Exception:
                # cert is a binary der encoded object
                cert_raw = b64_encode(self.logger, cert)

            # lookup REST-PATH of issuing CA
            ca_dic = self._ca_get_properties("name", self.ca_name)
            if "href" in ca_dic:
                error, cert_bundle = self._trigger_bundle_build(cert_raw, ca_dic)
            else:
                error = "Cannot find CA"
        else:
            error = "No payload given"

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)
